import pytest
import numpy as np
import dask.array as da
import numpy.testing as npt
import h5py
import os
import uuid
from pyxrf.core.fitting import fit_spectrum

from skbeam.core.fitting.background import snip_method

from pyxrf.core.map_processing import (
    dask_client_create,
    TerminalProgressBar,
    wait_and_display_progress,
    _compute_optimal_chunk_size,
    _chunk_numpy_array,
    _array_numpy_to_dask,
    RawHDF5Dataset,
    prepare_xrf_map,
    _prepare_xrf_mask,
    compute_total_spectrum,
    compute_total_spectrum_and_count,
    _fit_xrf_block,
    fit_xrf_map,
    _compute_roi,
    compute_selected_rois,
    snip_method_numba,
)

from pyxrf.core.tests.test_fitting import DataForFittingTest

import logging

logger = logging.getLogger(__name__)


def test_dask_client_create(tmpdir):
    """
    `dask_client_create` is a trivial function that instantiates Dask client
    in uniform way throughout the program. We test that we can pass addition
    kwargs to the object constructor or override default parameters.
    """
    # Set current directory. This is not required to create Dask client, but
    #   we will check to make sure that the directory for temporary files is
    #   not created in the current directory.
    os.chdir(tmpdir)
    dask_worker_space_path = os.path.join(tmpdir, "dask-worker-space")

    # Set the number of workers to some strange number (11 is unusual)
    #   (pass another kwarg in addition to default)
    client = dask_client_create(n_workers=11)
    n_workers = len(client.scheduler_info()["workers"])
    assert n_workers == 11, "The number of workers was set incorrectly"
    client.close()

    assert not os.path.exists(dask_worker_space_path), "Temporary directory was created in the current directory"

    # Disable multiprocessing: client is expected to have a single worker
    #   (replace the default value of the parameter)
    client = dask_client_create(processes=False)
    n_workers = len(client.scheduler_info()["workers"])
    assert n_workers == 1, "Dask client is expected to have one worker"
    client.close()

    assert not os.path.exists(dask_worker_space_path), "Temporary directory was created in the current directory"


def test_TerminalProgressBar():
    """Basic functionality of `TerminalProgressBar`"""

    # We just run the sequence of instructions. I didn't find a way to check the output,
    #   since 'progress' package requires TTY terminal. Otherwise output is blocked
    title = "Monitor progress"
    pbar = TerminalProgressBar(title)
    pbar.start()
    for n in range(10):
        pbar(n * 10)  # Go from 0 to 90%
    pbar.finish()  # This should set it to 100%


class _SampleProgressBar:
    """
    Progress bar class convenient for testing of the functions that
    display progress bar.
    """

    def __init__(self, title):
        self.title = title

        # The list of strings, which we expect to see in the output (capsys out)
        #   Used only for testing purposes only.
        self._expected_output = [
            f"Starting progress bar: {self.title}",
            "Percent completed: 100.0",
            f"Finished: {self.title}",
        ]

    def start(self):
        print(f"Starting progress bar: {self.title}")

    def __call__(self, percent_completed):
        print(f"Percent completed: {percent_completed}")

    def finish(self):
        print(f"Finished: {self.title}")

    def _check_output(self, capsys_out):
        # For testing purposes only, not needed for actual progress bar implementation
        for s in self._expected_output:
            assert (
                s in capsys_out
            ), f"Expected string {s} is missing in the progress bar object output:\n{capsys_out}"


# fmt: off
@pytest.mark.parametrize("progress_bar", [
    _SampleProgressBar("Monitoring progress"),
    TerminalProgressBar("Monitoring progress: "),
    None])
# fmt: on
def test_wait_and_display_progress(progress_bar, capsys):
    """Basic test for the function `wait_and_display_progress`"""

    # There is no way to monitor the output (no TTY device -> no output is generated)
    # So we just run a typical sequence of commands and make sure it doesn't crash

    client = dask_client_create()
    data = da.random.random(size=(100, 100), chunks=(10, 10))
    sm_fut = da.sum(data, axis=0).persist(scheduler=client)

    # Call the progress monitor
    wait_and_display_progress(sm_fut, progress_bar)

    sm = sm_fut.compute(scheduler=client)
    client.close()

    # Just in case check that the computations were done correctly.
    sm_expected = np.sum(data.compute(), axis=0)
    npt.assert_array_almost_equal(sm, sm_expected, err_msg="Computations are incorrect")

    # Make sure that the output contains all necessary components
    if isinstance(progress_bar, _SampleProgressBar):
        captured = capsys.readouterr()
        progress_bar._check_output(captured.out)


def test_RawHDF5Dataset(tmpdir):
    """Class RawHDF5Dataset"""
    dir_list, fln, dset = ("dir1", "dir2"), "tmp.txt", "/some/dataset"
    file_dir = os.path.join(tmpdir, *dir_list)
    os.makedirs(file_dir, exist_ok=True)
    os.chdir(file_dir)
    # Create file
    with open(fln, "w"):
        pass

    # 'fln' should be converted to absolute
    # 'shape' is set to some arbitrary valid tuple
    rds = RawHDF5Dataset(fln, dset, shape=(10, 50))

    assert os.path.isabs(rds.abs_path), "Path was not successfully converted to absolute path"
    # If file exists, then the file name was correctly set
    assert os.path.exists(rds.abs_path), "Path was not correctly set"
    assert rds.dset_name == dset, "Dataset name was not currectly set"
    assert rds.shape == (10, 50), "Attribute 'shape' was not set correctly"


# fmt: off
@pytest.mark.parametrize("data_chunksize, chunk_optimal, n_pixels, data_shape", [
    ((3, 5), (12, 10), 100, (20, 20)),
    ((5, 3), (10, 12), 100, (20, 20)),
    ((2, 5), (10, 10), 100, (20, 20)),
    ((11, 13), (11, 13), 100, (20, 20)),
    # The case of the stretched array
    ((2, 2), (6, 2), 10, (50, 2)),  # Stretched vertically
    ((2, 2), (2, 6), 10, (2, 50)),  # Stretched horizontally
    # Chunk size exceeds the size of the array: processed as one big chunk
    ((25, 25), (20, 20), 400, (20, 20)),
    # The number of pixels in the chunk exceeds the size of the array: same
    ((25, 25), (20, 20), 500, (20, 20)),
    # The original chunk size allows to divide the array by 4 chunks
    #   in order to parallelize processing (4 is the default value, but
    #   may be set)
    ((5, 2), (10, 10), 400, (20, 20)),
    ((15, 15), (15, 15), 400, (20, 20)),
])
# fmt: on
def test_compute_optimal_chunk_size(data_chunksize, chunk_optimal, n_pixels, data_shape):
    """Basic functionality of the '_compute_optimal_chunk_size'"""

    # Call with kwargs
    res = _compute_optimal_chunk_size(
        chunk_pixels=n_pixels, data_chunksize=data_chunksize, data_shape=data_shape, n_chunks_min=4
    )
    assert res == chunk_optimal, "Computed optimal chunks size doesn't match the expected"

    # Call with args
    res = _compute_optimal_chunk_size(n_pixels, data_chunksize, data_shape, 4)
    assert res == chunk_optimal, "Computed optimal chunks size doesn't match the expected"


# fmt: off
@pytest.mark.parametrize("data_chunksize, data_shape", [
    ((3,), (20, 20)),
    ((3, 5, 7), (20, 20)),
    (5, (20, 20)),
    ((5, 3), (20,)),
    ((5, 3), (20, 20, 20)),
    ((5, 3), 20),
])
# fmt: on
def test_compute_optimal_chunk_size_fail(data_chunksize, data_shape):
    """Failing cases for `compute_optimal_chunk_size`"""
    with pytest.raises(ValueError, match="Unsupported value of parameter"):
        _compute_optimal_chunk_size(10, data_chunksize, data_shape, 4)


# fmt: off
@pytest.mark.parametrize("chunk_target, data_shape", [
    ((2, 2), (10, 10, 3)),  # 3D array (primary use case)
    ((2, 3), (9, 11, 3)),
    ((3, 2), (11, 9, 3)),
    ((3, 3), (3, 3, 2)),
    ((1, 1), (1, 1, 3)),
    ((1, 5), (1, 10, 3)),
    ((5, 1), (10, 1, 3)),
    ((3, 4), (2, 3, 2)),
    ((2, 2), (9, 11, 5)),
    ((2, 2), (9, 11)),  # 2D array
    ((2, 2), (9, 11, 5, 3)),  # 4D array
    ((2, 2), (9, 11, 5, 3, 4)),  # 5D array

])
# fmt: on
def test_chunk_numpy_array(chunk_target, data_shape):
    """Basic functionailty tests for '_chunk_xrf_map_numpy"""

    data = np.random.random(data_shape)
    data_dask = _chunk_numpy_array(data, chunk_target)

    chunksize_expected = tuple(
        [min(chunk_target[0], data_shape[0]), min(chunk_target[1], data_shape[1]), *data_shape[2:]]
    )

    assert data_dask.shape == data.shape, "The shape of the original and chunked array don't match"
    assert (
        data_dask.chunksize == chunksize_expected
    ), "The chunk size of the Dask array doesn't match the desired chunk size"
    npt.assert_array_equal(
        data_dask.compute(), data, err_msg="The chunked array is different from the original array"
    )


# fmt: off
@pytest.mark.parametrize("chunk_pixels, data_shape, res_chunk_size", [
    (4, (10, 10, 3), (2, 2, 3)),
    (5, (10, 10, 3), (2, 3, 3)),
    (5, (50, 1, 3), (5, 1, 3)),
    (5, (1, 50, 3), (1, 5, 3)),
    (4, (9, 11), (2, 2)),
    (4, (9, 11, 3, 4), (2, 2, 3, 4)),
    (4, (9, 11, 3, 4, 2), (2, 2, 3, 4, 2)),
])
# fmt: on
def test_array_numpy_to_dask1(chunk_pixels, data_shape, res_chunk_size):
    """Basic functionality of `_xrf_map_numpy_to_dask`"""

    data = np.random.random(size=data_shape)

    res = _array_numpy_to_dask(data, chunk_pixels)

    assert res.shape == data.shape, "The shape of the original and chunked array don't match"
    assert res.chunksize == res_chunk_size, "The chunk size of the Dask array doesn't match the desired chunk size"


def test_array_numpy_to_dask2():
    """Test if `chunk_pixels` is properly adjusted if it is large compared to the XRF map
    size. Explicitely set the minimum number of chunks."""

    data_shape = (8, 8, 2)
    chunk_pixels = 100  # Some arbitrary large number
    data = np.random.random(size=data_shape)
    res = _array_numpy_to_dask(data, chunk_pixels, n_chunks_min=16)
    res_chunk_size = (2, 2, 2)

    assert res.shape == data.shape, "The shape of the original and chunked array don't match"
    assert res.chunksize == res_chunk_size, "The chunk size of the Dask array doesn't match the desired chunk size"


# fmt: off
@pytest.mark.parametrize("data", [
    da.random.random((5, 3, 5)),  # Dask array is not accepted
    10.6,  # Wrong data type
    (5, 6, 7),
    "test",
    np.random.random((3,)),  # Wrong number of dimensions
])
# fmt: on
def test_array_numpy_to_dask_fail(data):
    """Wrong type of `data` array"""
    with pytest.raises(ValueError, match="Parameter 'data' must numpy array with at least 2 dimensions"):
        _array_numpy_to_dask(data, (5, 2))


def _create_xrf_data(data_dask, data_representation, tmpdir, *, chunked_HDF5=True):
    """Prepare data represented as numpy array, dask array (no change) or HDF5 dataset"""
    if data_representation == "numpy_array":
        data = data_dask.compute()
    elif data_representation == "dask_array":
        data = data_dask
    elif data_representation == "hdf5_file_dset":
        os.chdir(tmpdir)
        fln = f"test-{uuid.uuid4()}.h5"  # Include UUID in the file name
        dset_name = "level1/level2"
        with h5py.File(fln, "w") as f:
            # In this test all computations are performed using 'float64' precision,
            #   so we create the dataset with dtype="float64" for consistency.
            kwargs = {"shape": data_dask.shape, "dtype": "float64"}
            if chunked_HDF5:
                kwargs.update({"chunks": data_dask.chunksize})
            dset = f.create_dataset(dset_name, **kwargs)
            dset[:, :, :] = data_dask.compute()
        data = RawHDF5Dataset(fln, dset_name, shape=data_dask.shape)
    else:
        raise RuntimeError(
            f"Error in test parameter: unknown value of 'data_representation' = {data_representation}"
        )

    return data


def _create_xrf_mask(data_shape, apply_mask, select_area):
    """
    Generate a mask for testing of XRF dataset processing functions.

    Parameters
    ----------
    data_shape: tuple or list
        (ny, nx, ...) - the shape of XRF dataset. All dimensions except 0 and 1 are ignored
    apply_mask: bool
        True: generate random mask,
    select_area: bool
        True: select area of the XRF map for processing.

    Returns
    -------
    mask: ndarray(int) or None
        mask is ndarray with shape (ny, nx). Integer values: 0 - pixel inactive,
        1, 2 - pixel active. Note: any positive integer marks pixel as active, but only
        values of 1 and 2 are generated.
    selection: tuple or None
        selected area is (y0, x0, ny_sel, nx_sel)

    If `select_area==True`, then all pixels in `mask` outside the selected area
    are disabled.
    """
    if apply_mask:
        mask = np.random.randint(0, 2, data_shape[0:2])
        mask[3, 4] = 0  # Make sure that at least 1 element is zero
        mask[2, 4] = 1  # Make sure that at least 1 element is non-zero
    else:
        mask = None

    # Selected area (between two points, the second point is not included)
    if select_area:
        selection = (2, 3, 4, 2)  # (y0, x0, ny, nx)
    else:
        selection = None

    return mask, selection


# fmt: off
@pytest.mark.parametrize("data_representation", ["numpy_array", "dask_array", "hdf5_file_dset"])
# fmt: on
def test_prepare_xrf_data(tmpdir, data_representation):
    # Start with dask array
    data_shape = (7, 12, 20)
    data_dask = da.random.random(data_shape, chunks=(2, 3, 4))
    data_numpy = data_dask.compute()

    data = _create_xrf_data(data_dask, data_representation, tmpdir)
    data, file_obj = prepare_xrf_map(data, chunk_pixels=12, n_chunks_min=4)

    assert (
        data.chunksize[0] * data.chunksize[1] == 12
    ), f"Dataset was not properly chunked: data.chunksize={data.chunksize}"

    npt.assert_array_almost_equal(
        data.compute(), data_numpy, err_msg="Prepared dataset is different from the original"
    )


def test_prepare_xrf_data_hdf5_not_chunked(tmpdir):
    # Start with dask array
    data_shape = (7, 12, 20)
    data_dask = da.random.random(data_shape, chunks=(2, 3, 4))
    data_numpy = data_dask.compute()

    data = _create_xrf_data(data_dask, "hdf5_file_dset", tmpdir, chunked_HDF5=False)

    data, file_obj = prepare_xrf_map(data, chunk_pixels=12, n_chunks_min=4)

    assert (data.chunksize[0] == 7) and (
        data.chunksize[1] == 12
    ), f"Dataset was not properly chunked: data.chunksize={data.chunksize}"

    npt.assert_array_almost_equal(
        data.compute(), data_numpy, err_msg="Prepared dataset is different from the original"
    )


def test_prepare_xrf_data_fail():
    """Failing test for `_prepare_xrf_mask_fail`"""
    data = 50.0  # Just a number
    with pytest.raises(TypeError, match="Type of parameter 'data' is not supported"):
        prepare_xrf_map(data, chunk_pixels=12, n_chunks_min=4)


@pytest.mark.parametrize("apply_mask", [False, True])
@pytest.mark.parametrize("select_area", [False, True])
def test_prepare_xrf_mask(apply_mask, select_area):
    """Basic functionality of `_prepare_xrf_mask`"""
    data_shape = (7, 12, 20)
    data_dask = da.random.random(data_shape, chunks=(2, 3, 4))
    mask, selection = _create_xrf_mask(data_shape, apply_mask, select_area)

    # Apply selection to mask
    mask_expected = None
    if mask is not None:
        mask_expected = mask
    if selection:
        if mask_expected is None:
            mask_expected = np.ones(shape=data_shape[0:2])
        ny, nx = mask_expected.shape
        y0, x0 = selection[0], selection[1]
        y1, x1 = y0 + selection[2], x0 + selection[3]
        for y in range(ny):
            if (y < y0) or (y >= y1):
                mask_expected[y, :] = 0
        for x in range(nx):
            if (x < x0) or (x >= x1):
                mask_expected[:, x] = 0

    mask_prepared = _prepare_xrf_mask(data_dask, mask=mask, selection=selection)

    npt.assert_array_equal(mask_prepared, mask_expected, err_msg="The prepared mask is not equal to expected")


# fmt: off
@pytest.mark.parametrize("data_dask, mask, selection, msg", [
    (np.random.random((5, 5, 10)), np.random.randint(0, 3, size=(5, 5)), None,
     "Parameter 'data' must be a Dask array"),
    (da.random.random((5,)), np.random.randint(0, 3, size=(5, 5)), None,
     "Parameter 'data' must have at least 2 dimensions"),
    (da.random.random((5, 4, 10)), np.random.randint(0, 3, size=(5, 5)), None,
     "Dimensions 0 and 1 of parameters 'data' and 'mask' do not match"),
    (da.random.random((5, 5, 10)), np.random.randint(0, 3, size=(5, 5)), (1, 3, 4),
     "Parameter 'selection' must be iterable with 4 elements"),
])
# fmt: on
def test_prepare_xrf_mask_fail(data_dask, mask, selection, msg):
    """Failing cases of `_prepare_xrf_mask`"""
    with pytest.raises(TypeError, match=msg):
        _prepare_xrf_mask(data_dask, mask=mask, selection=selection)


@pytest.fixture(scope="class")
def _start_dask_client(request):

    # Setup
    client = dask_client_create()

    # Set class variables
    request.cls.client = client
    yield

    # Tear down (after all tests in the class are completed)
    client.close()


@pytest.mark.usefixtures("_start_dask_client")
class TestComputeTotalSpectrum:
    @pytest.mark.parametrize("data_representation", ["numpy_array", "dask_array", "hdf5_file_dset"])
    @pytest.mark.parametrize("apply_mask", [False, True])
    @pytest.mark.parametrize("select_area", [False, True])
    def test_compute_total_spectrum1(self, data_representation, apply_mask, select_area, tmpdir):
        """Using 'global' dask client to save time"""

        global_client = self.client

        # Start with dask array
        data_shape = (7, 12, 20)
        data_dask = da.random.random(data_shape, chunks=(2, 3, 4))
        # Using the same distributed scheduler doesn't work as expected
        data_numpy = data_dask.compute(scheduler="synchronous")

        mask, selection = _create_xrf_mask(data_shape, apply_mask, select_area)

        # Compute the expected result
        data_tmp = data_numpy
        if mask is not None:
            mask_conv = (mask > 0).astype(dtype=int)
            mask_conv = np.broadcast_to(np.expand_dims(mask_conv, axis=2), data_tmp.shape)
            data_tmp = data_tmp * mask_conv
        if selection is not None:
            y0, x0, ny, nx = selection
            data_tmp = data_tmp[y0 : y0 + ny, x0 : x0 + nx, :]
        total_spectrum_expected = np.sum(np.sum(data_tmp, axis=0), axis=0)

        data = _create_xrf_data(data_dask, data_representation, tmpdir)

        total_spectrum = compute_total_spectrum(
            data,
            selection=selection,
            mask=mask,
            chunk_pixels=12,
            # Also run all computations with the progress bar
            progress_bar=TerminalProgressBar("Monitoring progress: "),
            client=global_client,
        )

        npt.assert_array_almost_equal(
            total_spectrum, total_spectrum_expected, err_msg="Total spectrum was computed incorrectly"
        )


def test_compute_total_spectrum2(tmpdir):
    """Create an instance of Dask client in the 'compute_total_spectrum' function to test if it works"""

    # Start with dask array
    data_shape = (7, 12, 20)
    data_dask = da.random.random(data_shape, chunks=(2, 3, 4))

    data_numpy = data_dask.compute()
    total_spectrum_expected = np.sum(np.sum(data_numpy, axis=0), axis=0)

    data = _create_xrf_data(data_dask, "dask_array", tmpdir=tmpdir)

    # Run computations without the progress bar
    total_spectrum = compute_total_spectrum(data, chunk_pixels=12)

    npt.assert_array_almost_equal(
        total_spectrum, total_spectrum_expected, err_msg="Total spectrum was computed incorrectly"
    )


# fmt: off
@pytest.mark.parametrize("mask", [da.random.random((7, 12)), (7, 12), 20, "abcde"])
# fmt: on
def test_compute_total_spectrum_fail(mask):
    """Failing cases: incorrect type for the mask ndarray"""
    data_dask = da.random.random((7, 12, 20), chunks=(2, 3, 4))
    with pytest.raises(TypeError, match="Parameter 'mask' must be a numpy array or None"):
        compute_total_spectrum(data_dask, mask=mask, chunk_pixels=12)


@pytest.mark.usefixtures("_start_dask_client")
class TestComputeTotalSpectrumAndCount:

    # fmt: off
    @pytest.mark.parametrize("data_representation", ["numpy_array", "dask_array", "hdf5_file_dset"])
    @pytest.mark.parametrize("apply_mask", [False, True])
    @pytest.mark.parametrize("select_area", [False, True])
    # fmt: on
    def test_compute_total_spectrum_and_count1(self, data_representation, apply_mask, select_area, tmpdir):
        """Using 'global' dask client to save time"""

        global_client = self.client

        # Start with dask array
        data_shape = (7, 12, 20)
        data_dask = da.random.random(data_shape, chunks=(2, 3, 4))
        # Using the same distributed scheduler doesn't work as expected
        data_numpy = data_dask.compute(scheduler="synchronous")

        mask, selection = _create_xrf_mask(data_shape, apply_mask, select_area)

        # Compute the expected result
        data_tmp = data_numpy
        if mask is not None:
            mask_conv = (mask > 0).astype(dtype=int)
            mask_conv = np.broadcast_to(np.expand_dims(mask_conv, axis=2), data_tmp.shape)
            data_tmp = data_tmp * mask_conv
        if selection is not None:
            y0, x0, ny, nx = selection
            # Primitive simple algorithm to apply the selection (we need to keep all the pixels)
            for y in range(data_tmp.shape[0]):
                for x in range(data_tmp.shape[1]):
                    if (y < y0) or (y >= y0 + ny) or (x < x0) or (x >= x0 + nx):
                        data_tmp[y, x, :] = 0.0
        total_spectrum_expected = np.sum(np.sum(data_tmp, axis=0), axis=0)
        total_count_expected = np.sum(data_tmp, axis=2)

        data = _create_xrf_data(data_dask, data_representation, tmpdir)

        total_spectrum, total_count = compute_total_spectrum_and_count(
            data,
            selection=selection,
            mask=mask,
            chunk_pixels=12,
            # Also run all computations with the progress bar
            progress_bar=TerminalProgressBar("Monitoring progress: "),
            client=global_client,
        )

        npt.assert_array_almost_equal(
            total_spectrum, total_spectrum_expected, err_msg="Total spectrum was computed incorrectly"
        )

        npt.assert_array_almost_equal(
            total_count, total_count_expected, err_msg="Total count (map) was computed incorrectly"
        )


def test_compute_total_spectrum_and_count2(tmpdir):
    """Create an instance of Dask client in the 'compute_total_spectrum' function to test if it works"""

    # Start with dask array
    data_shape = (7, 12, 20)
    data_dask = da.random.random(data_shape, chunks=(2, 3, 4))

    data_numpy = data_dask.compute()
    total_spectrum_expected = np.sum(np.sum(data_numpy, axis=0), axis=0)
    total_count_expected = np.sum(data_numpy, axis=2)

    data = _create_xrf_data(data_dask, "dask_array", tmpdir=tmpdir)

    # Run computations without the progress bar
    total_spectrum, total_count = compute_total_spectrum_and_count(data, chunk_pixels=12)

    npt.assert_array_almost_equal(
        total_spectrum, total_spectrum_expected, err_msg="Total spectrum was computed incorrectly"
    )

    npt.assert_array_almost_equal(
        total_count, total_count_expected, err_msg="Total count (map) was computed incorrectly"
    )


# fmt: off
@pytest.mark.parametrize("mask", [da.random.random((7, 12)), (7, 12), 20, "abcde"])
# fmt: on
def test_compute_total_spectrum_and_count_fail(mask):
    """Failing cases: incorrect type for the mask ndarray"""
    data_dask = da.random.random((7, 12, 20), chunks=(2, 3, 4))
    with pytest.raises(TypeError, match="Parameter 'mask' must be a numpy array or None"):
        compute_total_spectrum_and_count(data_dask, mask=mask, chunk_pixels=12)


class _FitXRFMapTesting:
    """
    The class implements methods for testing of XRF fitting algorithm.
    Used for testing `_fit_xrf_block` and `fit_xrf_map` functions.
    See the respective tests for examples
    """

    def __init__(self, *, dataset_params, use_snip, add_pts_before, add_pts_after):
        self.use_snip = use_snip
        self.add_pts_before = add_pts_before
        self.add_pts_after = add_pts_after

        self.fitting_data = DataForFittingTest(**dataset_params)

        # 'spectra' has dimensions (n_spec_points, n_lines), which is correct
        self.spectra = self.fitting_data.spectra
        self.n_spectrum_points, self.n_lines = self.spectra.shape

        # 'data_tmp' has dimensions (n_spec_points, ny, nx), so it needs to be rearranged
        self.data_tmp = self.fitting_data.data_input

        # Add some small background if snip is used
        if self.use_snip:
            self.data_tmp += 1.0

        # We want to also add points at the beginning and the end of the spectra
        # The original data is filled with random values. Those values should be either
        #   overwritten by actual spectral data or ignored during fitting.
        self.data_input = np.random.random(
            size=(
                self.data_tmp.shape[1],
                self.data_tmp.shape[2],
                self.data_tmp.shape[0] + self.add_pts_before + self.add_pts_after,
            )
        )

        # Range of indices of the experimental spectrum that should be used for fitting
        #   (it contains the spectrum data). The rest of the points should be ignored
        ne_start = self.add_pts_before
        ne_stop = self.add_pts_before + self.data_tmp.shape[0]
        self.data_sel_indices = (ne_start, ne_stop)

        for ny in range(self.data_tmp.shape[1]):
            for nx in range(self.data_tmp.shape[2]):
                ne_start = self.add_pts_before
                ne_stop = self.add_pts_before + self.data_tmp.shape[0]
                self.data_input[ny, nx, ne_start:ne_stop] = self.data_tmp[:, ny, nx]

        # The snip parameters are set so that the snip width is about 10 points
        self.snip_param = {"e_offset": 0.0, "e_linear": 0.1, "e_quadratic": 0.0, "b_width": 1}

    def verify_fit_output(self, *, data_out, snip_param=None):

        assert data_out.shape == (
            self.data_tmp.shape[1],
            self.data_tmp.shape[2],
            self.n_lines + 4,
        ), f"The shape of 'data_out' is incorrect: data_out.shape={data_out.shape}"

        weights_estimated = data_out[:, :, 0 : self.n_lines]
        bg_sum = data_out[:, :, self.n_lines]

        if not self.use_snip:
            # No background
            self.fitting_data.validate_output_weights(np.moveaxis(weights_estimated, 2, 0), decimal=10)

            assert (bg_sum == 0.0).all(), (
                f"Baseline estimate is non-zero when snip method is disabled: \n"
                f"bg_sum = {bg_sum}\nmin(bg_sum) = {np.min(bg_sum)}\nmax(bg_sum) = {np.max(bg_sum)}"
            )
        else:
            # Background is present in the data. Here we repreat the procedure
            #   of background subtraction and fitting. Unfortunately, the test code
            #   is very similar to the computational code used
            assert snip_param is not None, "Test parameter `snip_param` must be provided if 'use_snip' is enabled"
            _data = np.moveaxis(self.data_tmp, 0, 2)
            bg_sel = np.zeros(shape=_data.shape)
            for ny in range(bg_sel.shape[0]):
                for nx in range(bg_sel.shape[1]):
                    bg = snip_method_numba(
                        _data[ny, nx, :],
                        snip_param["e_offset"],
                        snip_param["e_linear"],
                        snip_param["e_quadratic"],
                        width=snip_param["b_width"],
                    )
                    bg_sel[ny, nx, :] = bg

            _data_no_bg = _data - bg_sel
            weights_expected, rfactor, _ = fit_spectrum(_data_no_bg, self.spectra, axis=2, method="nnls")
            npt.assert_array_almost_equal(
                weights_estimated,
                weights_expected,
                err_msg="Estimated weights are not equal to expected (use_snip==True)",
            )

            bg_sum_expected = np.sum(bg_sel, axis=2)
            npt.assert_array_almost_equal(
                bg_sum, bg_sum_expected, err_msg="Baseline is estimated incorrectly (use_snip==True)"
            )

        # Let's trust, that R-factor is correctly inserted into
        #   the 'data_out' array. Computation of R-factor is tested elsewhere,

        # Verify if the total count for the whole spectra and the selected region
        #   was computed correctly
        sm_total = np.sum(self.data_input, axis=2)
        sm_sel = np.sum(self.data_tmp, axis=0)  # Sum over the original dataset
        npt.assert_array_almost_equal(
            data_out[:, :, self.n_lines + 2],
            sm_sel,
            err_msg="Total count for the selected region is computed incorrectly",
        )
        npt.assert_array_almost_equal(
            data_out[:, :, self.n_lines + 3], sm_total, err_msg="Total count is computed incorrectly"
        )

    def verify_roi_output(self, *, data_out, roi_dict, snip_param=None):
        """
        Verify computated ROI. Computation of ROI is repeated in this function.
        Call with the same value of `snip_param` as the one sent to ROI computing function.
        It may be different from autogenerated `self.snip_param`.

        Parameters
        ----------
        data_out: dict
            Dictionary: key - emission line, value - 2D array with ROI values
        roi_dict: dict
            Dictionary: key - emission line, value - tuple (left_val, right_val)
            Energy values are in keV.
        snip_param: dict
            Parameters for SNIP algorithm for background subtraction. See `self.snip_param`
            defined in the constructor for the example. The values are used for subtracting
            baseline and for finding ranges of indices to define bands.
        """

        e_offset = snip_param["e_offset"]
        e_linear = snip_param["e_linear"]

        # Already truncated data (containing only selected region
        _data = np.moveaxis(self.data_tmp, 0, 2)

        if self.use_snip:
            bg_sel = np.zeros(shape=_data.shape)
            for ny in range(bg_sel.shape[0]):
                for nx in range(bg_sel.shape[1]):
                    bg = snip_method_numba(
                        _data[ny, nx, :],
                        snip_param["e_offset"],
                        snip_param["e_linear"],
                        snip_param["e_quadratic"],
                        width=snip_param["b_width"],
                    )
                    bg_sel[ny, nx, :] = bg
            _data = _data - bg_sel

        data_expected = {}
        for eline in roi_dict.keys():
            vleft, vright = roi_dict[eline]
            n_left = int(round((vleft - e_offset) / e_linear)) - self.add_pts_before
            n_right = int(round((vright - e_offset) / e_linear)) - self.add_pts_before
            n_left = int(np.clip(n_left, a_min=0, a_max=_data.shape[2] - 1))
            n_right = int(np.clip(n_right, a_min=0, a_max=_data.shape[2] - 1))
            if n_right > n_left:
                data_expected[eline] = np.sum(_data[:, :, n_left:n_right], axis=2)
            else:
                data_expected[eline] = np.zeros(shape=_data.shape[0:2])

        assert list(data_out.keys()) == list(
            data_expected.keys()
        ), "The list of output data keys is different from expected"

        for key in data_out.keys():
            npt.assert_array_almost_equal(
                data_out[key],
                data_expected[key],
                err_msg=f"Output ROI count for the key '{key}' is different from expected",
            )


# fmt: off
@pytest.mark.parametrize("dataset_params", [
    {"n_data_dimensions": (8, 1)},
    {"n_data_dimensions": (1, 8)},
    {"n_data_dimensions": (4, 4)},
    {"n_data_dimensions": (3, 5)},
])
@pytest.mark.parametrize("add_pts_before, add_pts_after", [(0, 0), (50, 100)])
@pytest.mark.parametrize("use_snip", [False, True])
# fmt: on
def test_fit_xrf_block(dataset_params, add_pts_before, add_pts_after, use_snip):

    ft = _FitXRFMapTesting(
        dataset_params=dataset_params,
        use_snip=use_snip,
        add_pts_before=add_pts_before,
        add_pts_after=add_pts_after,
    )

    data_out = _fit_xrf_block(
        ft.data_input,
        data_sel_indices=ft.data_sel_indices,
        matv=ft.spectra,
        snip_param=ft.snip_param,
        use_snip=use_snip,
    )

    ft.verify_fit_output(data_out=data_out, snip_param=ft.snip_param)


@pytest.mark.usefixtures("_start_dask_client")
class TestFitXRFMap:

    # fmt: off
    @pytest.mark.parametrize("data_representation", ["numpy_array", "dask_array", "hdf5_file_dset"])
    @pytest.mark.parametrize("dataset_params", [
        {"n_data_dimensions": (10, 10)},
        {"n_data_dimensions": (9, 11)},
        {"n_data_dimensions": (1, 100)},
        {"n_data_dimensions": (100, 1)},
    ])
    @pytest.mark.parametrize("add_pts", [(0, 0), (50, 100)])
    @pytest.mark.parametrize("use_snip", [False, True])
    # fmt: on
    def test_fit_xrf_map1(self, data_representation, dataset_params, add_pts, use_snip, tmpdir):
        """
        Basic functionality of `fit_xrf_map`.
        Tests are run using global Dask clients to inprove testing speed.
        """

        # Dask client object is used for multiple tests to save execution time
        global_client = self.client

        add_pts_before, add_pts_after = add_pts

        ft = _FitXRFMapTesting(
            dataset_params=dataset_params,
            use_snip=use_snip,
            add_pts_before=add_pts_before,
            add_pts_after=add_pts_after,
        )

        # Unfortunately 'data_input' is ndarray, but we need dask array to work with
        #   Select very small chunk size. This is not efficient, but works fine for testing.
        data_dask = _array_numpy_to_dask(ft.data_input, chunk_pixels=4, n_chunks_min=1)
        # Now create the dataset we need
        data = _create_xrf_data(data_dask, data_representation, tmpdir)

        # Run fitting
        data_out = fit_xrf_map(
            data,
            data_sel_indices=ft.data_sel_indices,
            matv=ft.spectra,
            snip_param=ft.snip_param,
            use_snip=use_snip,
            chunk_pixels=10,
            n_chunks_min=4,
            progress_bar=None,
            client=global_client,
        )

        ft.verify_fit_output(data_out=data_out, snip_param=ft.snip_param)


def test_fit_xrf_map2():
    """
    Basic functionality of `fit_xrf_map`.
    Tests are run using global Dask clients to inprove testing speed.
    """

    dataset_params = {"n_data_dimensions": (20, 20)}
    add_pts_before, add_pts_after = 15, 10
    use_snip = False

    ft = _FitXRFMapTesting(
        dataset_params=dataset_params,
        use_snip=use_snip,
        add_pts_before=add_pts_before,
        add_pts_after=add_pts_after,
    )

    # Just run the test with input data represented as numpy array
    data = ft.data_input

    # Run fitting
    data_out = fit_xrf_map(
        data,
        data_sel_indices=ft.data_sel_indices,
        matv=ft.spectra,
        snip_param=ft.snip_param,
        use_snip=use_snip,
        chunk_pixels=10,
        n_chunks_min=4,
        progress_bar=None,
        client=None,
    )

    ft.verify_fit_output(data_out=data_out, snip_param=ft.snip_param)


# fmt: off
@pytest.mark.parametrize("params, except_type, err_msg", [
    ({"data_sel_indices": 50}, TypeError,
     "Parameter 'data_sel_indices' must be tuple or list"),
    ({"data_sel_indices": (3, 10, 5)}, TypeError,
     "Parameter 'data_sel_indices' must contain two elements"),
    ({"data_sel_indices": [3]}, TypeError,
     "Parameter 'data_sel_indices' must contain two elements"),
    ({"data_sel_indices": (-1, 10)}, ValueError,
     "Some of the indices in 'data_sel_indices' are negative"),
    ({"data_sel_indices": [0, -10]}, ValueError,
     "Some of the indices in 'data_sel_indices' are negative"),
    ({"data_sel_indices": [3, 3]}, ValueError,
     "Parameter 'data_sel_indices' must select at least 1 element"),
    ({"data_sel_indices": [3, 2]}, ValueError,
     "Parameter 'data_sel_indices' must select at least 1 element"),
    ({"matv": np.zeros(shape=(2, 3, 5))}, TypeError,
     "Parameter 'matv' must be 2D ndarray"),
    ({"matv": np.zeros(shape=(50, 3)), "data_sel_indices": (5, 5 + 51)}, ValueError,
     "The number of selected points .* is not equal"),
    ({"snip_param": [1, 2, 3]}, TypeError,
     "Parameter 'snip_param' must be a dictionary"),
    ({"snip_param": {}, "use_snip": True}, TypeError,
     "Parameter 'snip_param' must a dictionary with keys"),
    ({"snip_param": {"e_offset": 0, "e_linear": 0.1, "e_quadratic": 0},
      "use_snip": True}, TypeError,
     "Parameter 'snip_param' must a dictionary with keys"),
    ({"data": np.zeros(shape=(10, 15, 100)), "matv": np.zeros(shape=(50, 3)),
      "data_sel_indices": (70, 70 + 50)}, ValueError,
     "Selection indices .* are outside the allowed range"),

])
# fmt: on
def test_fit_xrf_map_fail(params, except_type, err_msg):
    """Failing cases of `fit_xrf_map` (wrong input parameters)"""

    # Create a dataset first
    dataset_params = {"n_data_dimensions": (20, 20)}
    add_pts_before, add_pts_after = 15, 10
    use_snip = False

    ft = _FitXRFMapTesting(
        dataset_params=dataset_params,
        use_snip=use_snip,
        add_pts_before=add_pts_before,
        add_pts_after=add_pts_after,
    )

    kwargs = {
        "data": ft.data_input,
        "data_sel_indices": ft.data_sel_indices,
        "matv": ft.spectra,
        "snip_param": ft.snip_param,
        "use_snip": use_snip,
        "chunk_pixels": 10,
        "n_chunks_min": 4,
        "progress_bar": None,
        "client": None,
    }
    kwargs.update(params)

    # Run fitting
    with pytest.raises(except_type, match=err_msg):
        fit_xrf_map(**kwargs)


# fmt: off
@pytest.mark.parametrize("dataset_params", [
    {"n_data_dimensions": (8, 1)},
    {"n_data_dimensions": (1, 8)},
    {"n_data_dimensions": (4, 4)},
    {"n_data_dimensions": (3, 5)},
])
@pytest.mark.parametrize("add_pts_before, add_pts_after", [(0, 0), (50, 100)])
@pytest.mark.parametrize("use_snip", [False, True])
# fmt: on
def test_compute_roi(dataset_params, add_pts_before, add_pts_after, use_snip):
    """Basic functionality of `_compute_roi` function"""

    ft = _FitXRFMapTesting(
        dataset_params=dataset_params,
        use_snip=use_snip,
        add_pts_before=add_pts_before,
        add_pts_after=add_pts_after,
    )

    n_pts = ft.n_spectrum_points
    energy_min, energy_max = 3.1, 12.2  # Energy range for the 'selected' data, keV
    energy_step = (energy_max - energy_min) / (n_pts - 1)

    # 'e_offset' - the offset for the 0th element of the array, not the selected range
    snip_param = {
        "e_offset": energy_min - energy_step * add_pts_before,
        "e_linear": energy_step,
        "e_quadratic": 0,
        "b_width": 2.0,
    }

    roi_dict = {
        "roi-1": (2.5, 3.5),  # Outside the range
        "roi-2": (3.5, 4.8),
        "roi-3": (5.2, 7.4),
        "roi-4": (6.0, 6.0),  # No data points
        "roi-5": (6.0, 5.5),  # No data points
        "roi-6": (10.1, 15.0),  # Outside the range
    }

    roi_bands = [_ for _ in roi_dict.values()]

    # We don't use autogenerated 'ft.snip_param', because we want to use specific
    #   values to make sure that computations are done correctly.

    data_out = _compute_roi(
        ft.data_input,
        data_sel_indices=ft.data_sel_indices,
        roi_bands=roi_bands,
        snip_param=snip_param,
        use_snip=use_snip,
    )

    roi_keys = list(roi_dict.keys())
    assert data_out.shape == (*ft.data_input.shape[0:2], len(roi_keys)), "Output data has unexpected shape"

    # Convert data from numpy array to dictionary
    data_out = {roi_keys[_]: data_out[:, :, _] for _ in range(len(roi_keys))}
    # Verify the dictionary
    ft.verify_roi_output(data_out=data_out, roi_dict=roi_dict, snip_param=snip_param)


@pytest.mark.usefixtures("_start_dask_client")
class TestComputeSelectedROIs:

    # fmt: off
    @pytest.mark.parametrize("data_representation", ["numpy_array", "dask_array", "hdf5_file_dset"])
    @pytest.mark.parametrize("dataset_params", [
        {"n_data_dimensions": (10, 10)},
        {"n_data_dimensions": (9, 11)},
        {"n_data_dimensions": (1, 100)},
        {"n_data_dimensions": (100, 1)},
    ])
    @pytest.mark.parametrize("add_pts", [(0, 0), (50, 100)])
    @pytest.mark.parametrize("use_snip", [False, True])
    # fmt: on
    def test_compute_selected_rois1(self, data_representation, dataset_params, add_pts, use_snip, tmpdir):
        """
        Basic functionality of `compute_selected_rois`.
        Tests are run using global Dask clients to inprove testing speed.
        """

        # Dask client object is used for multiple tests to save execution time
        global_client = self.client

        add_pts_before, add_pts_after = add_pts

        ft = _FitXRFMapTesting(
            dataset_params=dataset_params,
            use_snip=use_snip,
            add_pts_before=add_pts_before,
            add_pts_after=add_pts_after,
        )

        # Unfortunately 'data_input' is ndarray, but we need dask array to work with
        #   Select very small chunk size. This is not efficient, but works fine for testing.
        data_dask = _array_numpy_to_dask(ft.data_input, chunk_pixels=4, n_chunks_min=1)
        # Now create the dataset we need
        data = _create_xrf_data(data_dask, data_representation, tmpdir)

        n_pts = ft.n_spectrum_points
        energy_min, energy_max = 3.1, 12.2  # Energy range for the 'selected' data, keV
        energy_step = (energy_max - energy_min) / (n_pts - 1)

        roi_dict = {
            "roi-1": (2.5, 3.5),  # Outside the range
            "roi-2": (3.5, 4.8),
            "roi-3": (5.2, 7.4),
            "roi-4": (6.0, 6.0),  # No data points
            "roi-5": (6.0, 5.5),  # No data points
            "roi-6": (10.1, 15.0),  # Outside the range
        }

        # 'e_offset' - the offset for the 0th element of the array, not the selected range
        snip_param = {
            "e_offset": energy_min - energy_step * add_pts_before,
            "e_linear": energy_step,
            "e_quadratic": 0,
            "b_width": 2.0,
        }

        # We don't use autogenerated 'ft.snip_param', because we want to use specific
        #   values to make sure that computations are done correctly.

        data_out = compute_selected_rois(
            data,
            data_sel_indices=ft.data_sel_indices,
            roi_dict=roi_dict,
            snip_param=snip_param,
            use_snip=use_snip,
            client=global_client,
        )

        # Verify the dictionary
        ft.verify_roi_output(data_out=data_out, roi_dict=roi_dict, snip_param=snip_param)


def test_compute_selected_rois2():
    """
    Basic functionality of `compute_selected_rois`.
    Tests are run using global Dask clients to inprove testing speed.
    """

    dataset_params = {"n_data_dimensions": (20, 20)}
    add_pts_before, add_pts_after = 15, 10
    use_snip = False

    ft = _FitXRFMapTesting(
        dataset_params=dataset_params,
        use_snip=use_snip,
        add_pts_before=add_pts_before,
        add_pts_after=add_pts_after,
    )

    # Just run the test with input data represented as numpy array
    data = ft.data_input

    n_pts = ft.n_spectrum_points
    energy_min, energy_max = 3.1, 12.2  # Energy range for the 'selected' data, keV
    energy_step = (energy_max - energy_min) / (n_pts - 1)

    roi_dict = {
        "roi-1": (2.5, 3.5),  # Outside the range
        "roi-2": (3.5, 4.8),
        "roi-3": (5.2, 7.4),
        "roi-4": (6.0, 6.0),  # No data points
        "roi-5": (6.0, 5.5),  # No data points
        "roi-6": (10.1, 15.0),  # Outside the range
    }

    # 'e_offset' - the offset for the 0th element of the array, not the selected range
    snip_param = {
        "e_offset": energy_min - energy_step * add_pts_before,
        "e_linear": energy_step,
        "e_quadratic": 0,
        "b_width": 2.0,
    }

    # We don't use autogenerated 'ft.snip_param', because we want to use specific
    #   values to make sure that computations are done correctly.

    data_out = compute_selected_rois(
        data, data_sel_indices=ft.data_sel_indices, roi_dict=roi_dict, snip_param=snip_param, use_snip=use_snip
    )

    # Verify the dictionary
    ft.verify_roi_output(data_out=data_out, roi_dict=roi_dict, snip_param=snip_param)


# fmt: off
@pytest.mark.parametrize("params, except_type, err_msg", [
    ({"data_sel_indices": 50}, TypeError,
     "Parameter 'data_sel_indices' must be tuple or list"),
    ({"data_sel_indices": (3, 10, 5)}, TypeError,
     "Parameter 'data_sel_indices' must contain two elements"),
    ({"data_sel_indices": [3]}, TypeError,
     "Parameter 'data_sel_indices' must contain two elements"),
    ({"data_sel_indices": (-1, 10)}, ValueError,
     "Some of the indices in 'data_sel_indices' are negative"),
    ({"data_sel_indices": [0, -10]}, ValueError,
     "Some of the indices in 'data_sel_indices' are negative"),
    ({"data_sel_indices": [3, 3]}, ValueError,
     "Parameter 'data_sel_indices' must select at least 1 element"),
    ({"data_sel_indices": [3, 2]}, ValueError,
     "Parameter 'data_sel_indices' must select at least 1 element"),
    ({"snip_param": [1, 2, 3]}, TypeError,
     "Parameter 'snip_param' must be a dictionary"),
    ({"snip_param": None}, TypeError,
     "Parameter 'snip_param' must be a dictionary"),
    ({"snip_param": {}}, TypeError,
     "Parameter 'snip_param' must a dictionary with keys"),
    ({"snip_param": {"e_offset": 0, "e_linear": 0.1, "e_quadratic": 0},
      "use_snip": True}, TypeError,
     "Parameter 'snip_param' must a dictionary with keys"),
    ({"data": np.zeros(shape=(10, 15, 100)),
      "data_sel_indices": (70, 70 + 50)}, ValueError,
     "Selection indices .* are outside the allowed range"),
])
# fmt: on
def test_compute_selected_rois_fail(params, except_type, err_msg):
    """Failing cases of `compute_selected_rois` (wrong input parameters)"""

    # Create a dataset first
    dataset_params = {"n_data_dimensions": (20, 20)}
    add_pts_before, add_pts_after = 15, 10
    use_snip = False

    ft = _FitXRFMapTesting(
        dataset_params=dataset_params,
        use_snip=use_snip,
        add_pts_before=add_pts_before,
        add_pts_after=add_pts_after,
    )

    # Just run the test with input data represented as numpy array
    data = ft.data_input

    # The particular values 'snip_param' are unimportant. Instead, the general
    #   structure of the parameter is tested. So we just use the autogenerated
    #   `ft.snip_param`, which is sometimes overriddent by the test parameter

    roi_dict = {
        "roi-1": (2.5, 3.5),  # Outside the range
        "roi-2": (3.5, 4.8),
        "roi-3": (5.2, 7.4),
        "roi-4": (6.0, 6.0),  # No data points
        "roi-5": (6.0, 5.5),  # No data points
        "roi-6": (10.1, 15.0),  # Outside the range
    }

    kwargs = {
        "data": data,
        "data_sel_indices": ft.data_sel_indices,
        "roi_dict": roi_dict,
        "snip_param": ft.snip_param,
        "use_snip": use_snip,
        "chunk_pixels": 10,
        "n_chunks_min": 4,
        "progress_bar": None,
        "client": None,
    }
    kwargs.update(params)

    # Run fitting
    with pytest.raises(except_type, match=err_msg):
        compute_selected_rois(**kwargs)


def test_snip_method_numba():
    """
    Compare the output of `snip_method_numba` with the output produced
    by `snip_method` to make sure the functions are equivalent.
    """

    dataset_params = {"n_data_dimensions": (20, 20)}
    add_pts_before, add_pts_after = 0, 0
    use_snip = True

    ft = _FitXRFMapTesting(
        dataset_params=dataset_params,
        use_snip=use_snip,
        add_pts_before=add_pts_before,
        add_pts_after=add_pts_after,
    )

    # Just run the test with input data represented as numpy array
    data = ft.data_input

    width = 0.5

    for ny in range(data.shape[0]):
        for nx in range(data.shape[1]):

            spec_sel = data[ny, nx, :]

            bg = snip_method_numba(spec_sel, 0, 0.01, 0, width=width)
            bg_expected = snip_method(spec_sel, 0, 0.01, 0, width=width)

            npt.assert_array_almost_equal(
                bg, bg_expected, err_msg=f"Background estimates don't match for the pixel ({ny}, {nx})"
            )



from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QComboBox, QCheckBox, QPushButton
from qtpy.QtCore import Signal, Slot

from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)

from .useful_widgets import set_tooltip, global_gui_variables

import logging

logger = logging.getLogger(__name__)


class PlotXrfMaps(QWidget):

    signal_maps_dataset_selection_changed = Signal()
    signal_maps_norm_changed = Signal()

    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        self._dataset_list = []  # The list of datasets ('combo_select_dataset')
        self._scaler_list = []  # The list of scalers ('combo_normalization')

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        # Reference to the main window. The main window will hold
        #   references to all non-modal windows that could be opened
        #   from multiple places in the program.
        self.ref_main_window = self.gui_vars["ref_main_window"]

        self.combo_select_dataset = QComboBox()
        self.combo_select_dataset.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.combo_select_dataset.currentIndexChanged.connect(self.combo_select_dataset_current_index_changed)

        self.combo_normalization = QComboBox()
        self.combo_normalization.currentIndexChanged.connect(self.combo_normalization_current_index_changed)

        self.pb_image_wizard = QPushButton("Image Wizard ...")
        self.pb_image_wizard.clicked.connect(self.pb_image_wizard_clicked)

        # self.pb_quant_settings = QPushButton("Quantitative ...")

        self.cb_interpolate = QCheckBox("Interpolate")
        self.cb_interpolate.setChecked(self.gpc.get_maps_grid_interpolate())
        self.cb_interpolate.toggled.connect(self.cb_interpolate_toggled)

        self.cb_scatter_plot = QCheckBox("Scatter plot")
        self.cb_scatter_plot.setChecked(self.gpc.get_maps_show_scatter_plot())
        self.cb_scatter_plot.toggled.connect(self.cb_scatter_plot_toggled)

        self.cb_quantitative = QCheckBox("Quantitative")
        self.cb_quantitative.setChecked(self.gpc.get_maps_quant_norm_enabled())
        self.cb_quantitative.toggled.connect(self.cb_quantitative_toggled)

        self.combo_color_scheme = QComboBox()
        # TODO: make color schemes global
        self._color_schemes = ("viridis", "jet", "bone", "gray", "Oranges", "hot")
        self.combo_color_scheme.addItems(self._color_schemes)
        self.combo_color_scheme.setCurrentIndex(self._color_schemes.index(self.gpc.get_maps_color_opt()))
        self.combo_color_scheme.currentIndexChanged.connect(self.combo_color_scheme_current_index_changed)

        self.combo_linear_log = QComboBox()
        self._linear_log_values = ["Linear", "Log"]
        self.combo_linear_log.addItems(["Linear", "Log"])
        scale_opt = self.gpc.get_maps_scale_opt()
        ind = self._linear_log_values.index(scale_opt)
        self.combo_linear_log.setCurrentIndex(ind)
        self.combo_linear_log.currentIndexChanged.connect(self.combo_linear_log_current_index_changed)

        self.combo_pixels_positions = QComboBox()
        self._pix_pos_values = ["Pixels", "Positions"]
        self.combo_pixels_positions.addItems(self._pix_pos_values)
        self.combo_pixels_positions.setCurrentIndex(self._pix_pos_values.index(self.gpc.get_maps_pixel_or_pos()))
        self.combo_pixels_positions.currentIndexChanged.connect(self.combo_pixels_positions_current_index_changed)

        self.mpl_canvas = FigureCanvas(self.gpc.img_model_adv.fig)
        self.mpl_toolbar = NavigationToolbar(self.mpl_canvas, self)

        # Keep layout without change when canvas is hidden (invisible)
        sp_retain = self.mpl_canvas.sizePolicy()
        sp_retain.setRetainSizeWhenHidden(True)
        self.mpl_canvas.setSizePolicy(sp_retain)

        vbox = QVBoxLayout()
        hbox = QHBoxLayout()
        hbox.addWidget(self.combo_select_dataset)
        hbox.addStretch(4)
        hbox.addWidget(self.pb_image_wizard)
        hbox.addStretch(1)
        hbox.addWidget(self.cb_interpolate)
        hbox.addWidget(self.cb_scatter_plot)
        vbox.addLayout(hbox)
        hbox = QHBoxLayout()
        hbox.addWidget(self.combo_normalization)
        hbox.addStretch(3)
        hbox.addWidget(self.cb_quantitative)
        hbox.addStretch(1)
        hbox.addWidget(self.combo_linear_log)
        hbox.addWidget(self.combo_pixels_positions)
        hbox.addWidget(self.combo_color_scheme)
        vbox.addLayout(hbox)
        vbox.addWidget(self.mpl_toolbar)
        vbox.addWidget(self.mpl_canvas)
        self.setLayout(vbox)

        self._set_tooltips()

    def _set_tooltips(self):
        set_tooltip(self.combo_select_dataset, "Select <b>dataset</b>.")
        set_tooltip(self.combo_normalization, "Select <b>scaler</b> for normalization of displayed XRF maps.")
        set_tooltip(
            self.pb_image_wizard,
            "Open the window with tools for <b>selection and configuration</b> the displayed XRF maps.",
        )
        set_tooltip(self.cb_interpolate, "Interpolate coordinates to <b>uniform grid</b>.")
        set_tooltip(self.cb_scatter_plot, "Display <b>scatter plot</b>.")
        set_tooltip(
            self.cb_quantitative,
            "Normalize the displayed XRF maps using loaded <b>Quantitative Calibration</b> data.",
        )
        set_tooltip(self.combo_color_scheme, "Select <b>color scheme</b>")
        set_tooltip(
            self.combo_linear_log,
            "Switch between <b>linear</b> and <b>logarithmic</b> scale "
            "for plotting of XRF Map pixel <b>intensity</b>.",
        )
        set_tooltip(
            self.combo_pixels_positions, "Switch axes units between <b>pixels</b> and <b>positional units</b>."
        )

    def update_widget_state(self, condition=None):
        if condition == "tooltips":
            self._set_tooltips()
        self.mpl_toolbar.setVisible(self.gui_vars["show_matplotlib_toolbar"])

        # Hide Matplotlib canvas during computations
        state_compute = global_gui_variables["gui_state"]["running_computations"]
        self.mpl_canvas.setVisible(not state_compute)

    def pb_image_wizard_clicked(self):
        # Position the window in relation ot the main window (only when called once)
        pos = self.ref_main_window.pos()
        self.ref_main_window.wnd_image_wizard.position_once(pos.x(), pos.y())

        if not self.ref_main_window.wnd_image_wizard.isVisible():
            self.ref_main_window.wnd_image_wizard.show()
        self.ref_main_window.wnd_image_wizard.activateWindow()

    def cb_scatter_plot_toggled(self, state):
        self.gpc.set_maps_show_scatter_plot(state)
        self.cb_interpolate.setVisible(not state)
        self.combo_pixels_positions.setVisible(not state)

    def cb_quantitative_toggled(self, state):
        self.gpc.set_maps_quant_norm_enabled(state)
        self.signal_maps_norm_changed.emit()

    def combo_select_dataset_current_index_changed(self, index):
        self.gpc.set_maps_selected_dataset(index + 1)
        self.signal_maps_dataset_selection_changed.emit()

    @Slot()
    def combo_select_dataset_update_current_index(self):
        index = self.gpc.get_maps_selected_dataset()
        self.combo_select_dataset.setCurrentIndex(index - 1)

    def combo_normalization_current_index_changed(self, index):
        self.gpc.set_maps_scaler_index(index)
        self.signal_maps_norm_changed.emit()

    def combo_linear_log_current_index_changed(self, index):
        self.gpc.set_maps_scale_opt(self._linear_log_values[index])

    def combo_color_scheme_current_index_changed(self, index):
        self.gpc.set_maps_color_opt(self._color_schemes[index])

    def combo_pixels_positions_current_index_changed(self, index):
        self.gpc.set_maps_pixel_or_pos(self._pix_pos_values[index])

    def cb_interpolate_toggled(self, state):
        self.gpc.set_maps_grid_interpolate(state)

    @Slot()
    def slot_update_dataset_info(self):
        self._update_datasets()
        self._update_scalers()
        self.cb_quantitative.setChecked(self.gpc.get_maps_quant_norm_enabled())

    def _update_datasets(self):
        dataset_list, dset_sel = self.gpc.get_maps_dataset_list()
        self._dataset_list = dataset_list.copy()
        self.combo_select_dataset.clear()
        self.combo_select_dataset.addItems(self._dataset_list)
        # No item should be selected if 'dset_sel' is 0
        self.combo_select_dataset.setCurrentIndex(dset_sel - 1)

    def _update_scalers(self):
        scalers, scaler_sel = self.gpc.get_maps_scaler_list()
        self._scaler_list = ["Normalize by ..."] + scalers
        self.combo_normalization.clear()
        self.combo_normalization.addItems(self._scaler_list)
        self.combo_normalization.setCurrentIndex(scaler_sel)



from __future__ import absolute_import, division, print_function
import math
import numpy as np
import matplotlib
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.collections import BrokenBarHCollection
import matplotlib.ticker as mticker
from matplotlib.colors import LogNorm
from enum import Enum
from mpl_toolkits.axes_grid1 import ImageGrid

from atom.api import Atom, Str, observe, Typed, Int, List, Dict, Float, Bool

from skbeam.core.fitting.xrf_model import K_TRANSITIONS, L_TRANSITIONS, M_TRANSITIONS
from skbeam.fluorescence import XrfElement as Element

from ..core.xrf_utils import get_eline_parameters

import logging

logger = logging.getLogger(__name__)


def get_color_name():

    # usually line plot will not go beyond 10
    first_ten = [
        "indigo",
        "maroon",
        "green",
        "darkblue",
        "darkgoldenrod",
        "blue",
        "darkcyan",
        "sandybrown",
        "black",
        "darkolivegreen",
    ]

    # Avoid red color, as those color conflict with emission lines' color.
    nonred_list = [
        v
        for v in matplotlib.colors.cnames.keys()
        if "pink" not in v and "fire" not in v and "sage" not in v and "tomato" not in v and "red" not in v
    ]
    return first_ten + nonred_list + list(matplotlib.colors.cnames.keys())


class PlotTypes(Enum):
    LINLOG = 0
    LINEAR = 1


class EnergyRangePresets(Enum):
    SELECTED_RANGE = 0
    FULL_SPECTRUM = 1


class MapTypes(Enum):
    LINEAR = 0
    LOG = 1


class MapAxesUnits(Enum):
    PIXELS = 0
    POSITIONS = 1


class LinePlotModel(Atom):
    """
    This class performs all the required line plots.

    Attributes
    ----------
    data : array
        Experimental data
    _fit : class object
        Figure object from matplotlib
    _ax : class object
        Axis object from matplotlib
    _canvas : class object
        Canvas object from matplotlib
    element_id : int
        Index of element
    parameters : `atom.List`
        A list of `Parameter` objects, subclassed from the `Atom` base class.
        These `Parameter` objects hold all relevant xrf information
    elist : list
        Emission energy and intensity for given element
    plot_opt : int
        Linear or log plot
    total_y : dict
        Results for k lines
    total_l : dict
        Results for l and m lines
    prefit_x : array
        X axis with limited range
    plot_title : str
        Title for plotting
    fit_x : array
        x value for fitting
    fit_y : array
        fitted data
    plot_type_names : list
        linear or log plot
    max_v : float
        max value of data array
    incident_energy : float
        in KeV
    param_model : Typed(object)
        Reference to ParamModel object
    """

    # data = Typed(object)  # Typed(np.ndarray)
    exp_data_label = Str("experiment")

    number_pts_to_show = Int(3000)  # The number of spectrum point to show

    # -------------------------------------------------------------
    # Preview plot (raw experimental spectra)
    _fig_preview = Typed(Figure)
    _ax_preview = Typed(Axes)
    _lines_preview = List()
    _bahr_preview = Typed(BrokenBarHCollection)

    plot_type_preview = Typed(PlotTypes)
    energy_range_preview = Typed(EnergyRangePresets)

    min_v_preview = Float()
    max_v_preview = Float()
    min_e_preview = Float()
    max_e_preview = Float()

    # -----------------------------------------------------------
    # Preview of Total Count Maps
    _fig_maps = Typed(Figure)
    map_type_preview = Typed(MapTypes)
    map_axes_units_preview = Typed(MapAxesUnits)
    map_scatter_plot = Bool(False)

    map_preview_color_scheme = Str("viridis")
    map_preview_range_low = Float(-1)
    map_preview_range_high = Float(-1)
    # ------------------------------------------------------------

    _fig = Typed(Figure)
    _ax = Typed(Axes)
    _canvas = Typed(object)
    plot_fit_x_min = Float(0)  # The variables are used to store x_min and x_max for the current plot
    plot_fit_x_max = Float(0)
    element_id = Int(0)
    elist = List()
    scale_opt = Int(0)
    # total_y = Dict()
    # total_l = Dict()
    # total_m = Dict()
    # total_pileup = Dict()

    prefit_x = Typed(object)
    plot_title = Str()
    # fit_x = Typed(np.ndarray)
    # fit_y = Typed(np.ndarray)
    # residual = Typed(np.ndarray)

    plot_type_names = List()
    max_v = Float()
    incident_energy = Float(12.0)

    energy_range_names = List()
    energy_range_fitting = Str()

    eline_obj = List()

    plot_exp_opt = Bool(False)
    plot_exp_obj = Typed(Line2D)
    show_exp_opt = Bool(False)  # Flag: show spectrum preview

    # Reference to artist responsible for displaying the selected range of energies on the plot
    plot_energy_barh = Typed(BrokenBarHCollection)
    t_bar = Typed(object)

    plot_exp_list = List()

    auto_fit_obj = List()
    show_autofit_opt = Bool()

    plot_fit_obj = List()  # Typed(Line2D)
    show_fit_opt = Bool(False)
    # fit_all = Typed(object)

    plot_style = Dict()

    roi_plot_dict = Dict()
    roi_dict = Typed(object)  # OrderedDict()

    log_range = List()
    linear_range = List()
    plot_escape_line = Int(0)
    emission_line_window = Bool(True)
    det_materials = Int(0)
    escape_e = Float(1.73998)
    limit_cut = Int()
    # prefix_name_roi = Str()
    # element_for_roi = Str()
    # element_list_roi = List()
    # roi_dict = Typed(object) #OrderedDict()

    # img_dict = Dict()
    # roi_result = Dict()

    # Reference to ParamModel object
    param_model = Typed(object)
    # Reference to FileIOModel object
    io_model = Typed(object)

    # Location of the vertical (mouse-selected) marker on the plot.
    # Value is in kev. Negative value - no marker is placed.
    vertical_marker_kev = Float(-1)
    # Reference to the respective Matplotlib artist
    line_vertical_marker = Typed(object)
    vertical_marker_is_visible = Bool(False)

    report_marker_state = Typed(object)

    def __init__(self, *, param_model, io_model):

        # Reference to ParamModel object
        self.param_model = param_model
        self.io_model = io_model

        # self.data = None

        self._fig = plt.figure()

        self._ax = self._fig.add_subplot(111)
        try:
            self._ax.set_axis_bgcolor("lightgrey")
        except AttributeError:
            self._ax.set_facecolor("lightgrey")

        self._ax.set_xlabel("Energy (keV)")
        self._ax.set_ylabel("Spectrum (Counts)")
        self._ax.grid(which="both")
        self._ax.set_yscale("log")
        self.plot_type_names = ["LinLog", "Linear"]

        self.energy_range_names = ["selected", "full"]
        self.energy_range_fitting = "selected"

        self._ax.autoscale_view(tight=True)
        self._ax.legend(loc=2)

        self._color_config()
        self._fig.tight_layout(pad=0.5)
        self.max_v = 1.0
        # when we calculate max value, data smaller than 500, 0.5 Kev, can be ignored.
        # And the last point of data is also huge, and should be cut off.
        self.limit_cut = 100
        # self._ax.margins(x=0.0, y=0.10)

        # --------------------------------------------------------------
        # Spectrum preview figure
        self._fig_preview = Figure()
        self.plot_type_preview = PlotTypes.LINLOG
        self.energy_range_preview = EnergyRangePresets.SELECTED_RANGE

        # --------------------------------------------------------------
        # Preview of Total Count Maps
        self._fig_maps = Figure()
        self.map_type_preview = MapTypes.LINEAR
        self.map_axes_units_preview = MapAxesUnits.PIXELS

    def _color_config(self):
        self.plot_style = {
            "experiment": {"color": "blue", "linestyle": "", "marker": ".", "label": self.exp_data_label},
            "background": {"color": "indigo", "marker": "+", "markersize": 1, "label": "background"},
            "emission_line": {"color": "black", "linewidth": 2},
            "roi_line": {"color": "red", "linewidth": 2},
            "k_line": {"color": "green", "label": "k lines"},
            "l_line": {"color": "magenta", "label": "l lines"},
            "m_line": {"color": "brown", "label": "m lines"},
            "compton": {"color": "darkcyan", "linewidth": 1.5, "label": "compton"},
            "elastic": {"color": "purple", "label": "elastic"},
            "escape": {"color": "darkblue", "label": "escape"},
            "pileup": {"color": "darkgoldenrod", "label": "pileup"},
            "userpeak": {"color": "orange", "label": "userpeak"},
            # 'auto_fit': {'color': 'black', 'label': 'auto fitted', 'linewidth': 2.5},
            "fit": {"color": "red", "label": "fit", "linewidth": 2.5},
            "residual": {"color": "black", "label": "residual", "linewidth": 2.0},
        }

    def plot_exp_data_update(self, change):
        """
        Observer function to be connected to the fileio model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : dict
            This is the dictionary that gets passed to a function
            with the @observe decorator
        """
        self.plot_exp_opt = False  # exp data for fitting
        self.show_exp_opt = False  # all exp data from different channels
        self.show_fit_opt = False

        # Reset currently selected element_id (mostly to reset GUI elements)
        self.element_id = 0

    def init_mouse_event(self):
        """Set up callback for mouse button-press event"""
        # Reference to the toolbar
        self.t_bar = self._fig.canvas.toolbar
        # Set callback for Button Press event
        self._fig.canvas.mpl_connect("button_press_event", self.canvas_onpress)

    def _update_canvas(self):
        # It may be sufficient to initialize the event only once, but at this point
        #   it seems to be the most reliable option. May be changed in the future.
        self.init_mouse_event()
        self.plot_vertical_marker()

        self._ax.legend(loc=2)
        try:
            self._ax.legend(framealpha=0.2).set_draggable(True)
        except AttributeError:
            self._ax.legend(framealpha=0.2)
        self._fig.tight_layout(pad=0.5)
        # self._ax.margins(x=0.0, y=0.10)

        # when we click the home button on matplotlib gui,
        # relim will remember the previously defined x range
        self._ax.relim(visible_only=True)
        self._fig.canvas.draw()

    def _update_ylimit(self):
        # manually define y limit, from experience
        self.log_range = [self.max_v * 1e-5, self.max_v * 2]
        # self.linear_range = [-0.3*self.max_v, self.max_v*1.2]
        self.linear_range = [0, self.max_v * 1.2]

    def exp_label_update(self, change):
        """
        Observer function to be connected to the fileio model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : dict
            This is the dictionary that gets passed to a function
            with the @observe decorator
        """
        self.exp_data_label = change["value"]
        self.plot_style["experiment"]["label"] = change["value"]

    # @observe('exp_data_label')
    # def _change_exp_label(self, change):
    #     if change['type'] == 'create':
    #         return
    #     self.plot_style['experiment']['label'] = change['value']

    @observe("parameters")
    def _update_energy(self, change):
        if "coherent_sct_energy" not in self.param_model.param_new:
            return
        self.incident_energy = self.param_model.param_new["coherent_sct_energy"]["value"]

    def set_energy_range_fitting(self, energy_range_name):
        if energy_range_name not in self.energy_range_names:
            raise ValueError(
                f"Unknown energy range name {energy_range_name}. Allowed names: {self.energy_range_names}"
            )
        self.energy_range_fitting = energy_range_name
        self.plot_experiment()

    def set_incident_energy(self, change):
        """
        The observer function that changes the value of incident energy
        and upper bound for fitted energy range. Should not be called directly.

        Parameters
        ----------

        change : dict
            ``change["value"]`` is the new value of incident energy
        """
        self.change_incident_energy(change["value"])

    def change_incident_energy(self, energy_new):
        """
        The function that perfroms the changes the value of incident energy
        and upper bound for fitted energy range.

        Parameters
        ----------

        incident_energy : float
            New value of incident energy
        """

        margin = 0.8  # Value by which the upper bound of the range used for fitting
        #               exceeds the incident energy. Selected for convenience, but
        #               is subject to change. This is the place to change it to take effect
        #               throughout the program.

        # Limit the number of decimal points for better visual presentation
        energy_new = round(energy_new, ndigits=6)
        # Change the value twice to ensure that all observer functions are called
        self.incident_energy = energy_new + 1.0  # Arbitrary number different from 'energy_new'
        self.incident_energy = energy_new
        if "coherent_sct_energy" in self.param_model.param_new:
            self.param_model.param_new["coherent_sct_energy"]["value"] = energy_new

        # Change the value twice to ensure that all observer functions are called
        self.param_model.energy_bound_high_buf = energy_new + 1.8  # Arbitrary number
        upper_bound = energy_new + margin
        # Limit the number of decimal points for better visual presentation
        upper_bound = round(upper_bound, ndigits=5)
        self.param_model.energy_bound_high_buf = upper_bound

    @observe("scale_opt")
    def _new_opt(self, change):
        self.log_linear_plot()
        self._update_canvas()

    def energy_bound_high_update(self, change):
        """Observer function for 'param_model.energy_bound_high_buf'"""
        if self.io_model.data is None:
            return
        self.exp_data_update({"value": self.io_model.data})
        self.plot_selected_energy_range_original(e_high=change["value"])
        self.plot_vertical_marker(e_high=change["value"])
        self._update_canvas()

    def energy_bound_low_update(self, change):
        """Observer function for 'param_model.energy_bound_low_buf'"""
        if self.io_model.data is None:
            return
        self.exp_data_update({"value": self.io_model.data})
        self.plot_selected_energy_range_original(e_low=change["value"])
        self.plot_vertical_marker(e_low=change["value"])
        self._update_canvas()

    def log_linear_plot(self):
        if self.plot_type_names[self.scale_opt] == "LinLog":
            self._ax.set_yscale("log")
            # self._ax.margins(x=0.0, y=0.5)
            # self._ax.autoscale_view(tight=True)
            # self._ax.relim(visible_only=True)
            self._ax.set_ylim(self.log_range)

        else:
            self._ax.set_yscale("linear")
            # self._ax.margins(x=0.0, y=0.10)
            # self._ax.autoscale_view(tight=True)
            # self._ax.relim(visible_only=True)
            self._ax.set_ylim(self.linear_range)

    def exp_data_update(self, change):
        """
        Observer function to be connected to the fileio model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : dict
            This is the dictionary that gets passed to a function
            with the @observe decorator
        """
        # TODO: This function does not change the data. Instead it is expected to
        #   perform a number of operation when data is changed.

        # self.data = change['value']
        if self.io_model.data is None:
            return

        e_range = self.energy_range_fitting
        e_range_full, e_range_selected = "full", "selected"
        if set([e_range_full, e_range_selected]) < set(self.energy_range_names):
            raise ValueError(
                f"Some names for energy range {(e_range_full, e_range_selected)} are not supported. "
                "Please report the error to the development team."
            )
        if e_range not in (e_range_full, e_range_selected):
            logger.error(
                f"Spectrum preview: Unknown option for the energy range: {e_range}\n"
                "Please report the error to the development team."
            )
            # This is not a critical error, so we still can proceed
            e_range = e_range_full

        if not self.param_model.param_new:
            return

        # The number of points in the displayed dataset
        n_dset_points = len(self.io_model.data)

        if e_range == e_range_selected:
            n_range_low, n_range_high = self.selected_range_indices(n_indexes=n_dset_points)
        else:
            n_range_low, n_range_high = 0, n_dset_points

        n_low = int(np.clip(n_range_low, a_min=0, a_max=n_dset_points - 1))
        n_high = int(np.clip(n_range_high, a_min=1, a_max=n_dset_points))

        # Find the maximum value (skip the first and last 'limit_cut' points of the dataset
        n1, n2 = max(self.limit_cut, n_low), min(n_dset_points - self.limit_cut, n_high)
        if n2 <= n1:  # This is just a precaution: it is expected that n_dset_points >> 2 * limit_cut
            n1, n2 = n_low, n_high
        self.max_v = float(np.max(self.io_model.data[n1:n2]))

        try:
            self.plot_exp_obj.remove()
            logger.debug("Previous experimental data is removed.")
        except AttributeError:
            logger.debug("No need to remove experimental data.")

        data_arr = self.io_model.data
        x_v = (
            self.param_model.param_new["e_offset"]["value"]
            + np.arange(n_low, n_high) * self.param_model.param_new["e_linear"]["value"]
            + np.arange(n_low, n_high) ** 2 * self.param_model.param_new["e_quadratic"]["value"]
        )

        data_arr = data_arr[n_low:n_high]

        (self.plot_exp_obj,) = self._ax.plot(
            x_v,
            data_arr,
            linestyle=self.plot_style["experiment"]["linestyle"],
            color=self.plot_style["experiment"]["color"],
            marker=self.plot_style["experiment"]["marker"],
            label=self.plot_style["experiment"]["label"],
        )

        # Rescale the plot along x-axis if needed
        x_min, x_max = x_v[0], x_v[-1]
        if (x_min != self.plot_fit_x_min) or (x_max != self.plot_fit_x_max):
            self.plot_fit_x_min = x_min
            self.plot_fit_x_max = x_max
            self._ax.set_xlim(x_min, x_max)

        self._update_ylimit()
        self.log_linear_plot()

        self._set_eline_select_controls()

        self.plot_selected_energy_range_original()

        # _show_hide_exp_plot is called to show or hide current plot based
        #           on the state of _show_exp_opt flag
        self._show_hide_exp_plot(self.show_exp_opt or self.plot_exp_opt)

    def _show_hide_exp_plot(self, plot_show):

        if self.io_model.data is None:
            return

        try:
            if plot_show:
                self.plot_exp_obj.set_visible(True)
                lab = self.plot_exp_obj.get_label()
                self.plot_exp_obj.set_label(lab.strip("_"))
            else:
                self.plot_exp_obj.set_visible(False)
                lab = self.plot_exp_obj.get_label()
                self.plot_exp_obj.set_label("_" + lab)

            self._update_canvas()
        except Exception:
            pass

    @observe("plot_exp_opt")
    def _new_exp_plot_opt(self, change):

        if self.io_model.data is None:
            return

        if change["type"] != "create":
            if change["value"]:
                self.plot_experiment()
            # _show_hide_exp_plot is already called inside 'plot_experiment()',
            #    but visibility flag was not used correctly. So we need to
            #    call it again.
            self._show_hide_exp_plot(change["value"])
            self._set_eline_select_controls()

    # @observe('show_exp_opt')
    # def _update_exp(self, change):
    #     if change['type'] != 'create':
    #         if change['value']:
    #             if len(self.plot_exp_list):
    #                 for v in self.plot_exp_list:
    #                     v.set_visible(True)
    #                     lab = v.get_label()
    #                     if lab != '_nolegend_':
    #                         v.set_label(lab.strip('_'))
    #         else:
    #             if len(self.plot_exp_list):
    #                 for v in self.plot_exp_list:
    #                     v.set_visible(False)
    #                     lab = v.get_label()
    #                     if lab != '_nolegend_':
    #                         v.set_label('_' + lab)
    #         self._update_canvas()

    @observe("show_fit_opt")
    def _update_fit(self, change):
        if change["type"] != "create":
            if change["value"]:
                for v in self.plot_fit_obj:
                    v.set_visible(True)
                    lab = v.get_label()
                    if lab != "_nolegend_":
                        v.set_label(lab.strip("_"))
            else:
                for v in self.plot_fit_obj:
                    v.set_visible(False)
                    lab = v.get_label()
                    if lab != "_nolegend_":
                        v.set_label("_" + lab)
            self._update_canvas()

    def plot_experiment(self):
        """
        PLot raw experiment data for fitting.
        """

        # Do nothing if no data is loaded
        if self.io_model.data is None:
            return

        data_arr = np.asarray(self.io_model.data)
        self.exp_data_update({"value": data_arr})

    def plot_vertical_marker(self, *, e_low=None, e_high=None):

        # It doesn't seem necessary to force the marker inside the selected range.
        #   It may be used for purposes that require to set it outside the range
        # self._vertical_marker_set_inside_range(e_low=e_low, e_high=e_high)

        x_v = (self.vertical_marker_kev, self.vertical_marker_kev)
        y_v = (-1e30, 1e30)  # This will cover the range of possible values of accumulated counts

        if self.line_vertical_marker:
            self._ax.lines.remove(self.line_vertical_marker)
            self.line_vertical_marker = None
        if self.vertical_marker_is_visible:
            (self.line_vertical_marker,) = self._ax.plot(x_v, y_v, color="blue")

    def set_plot_vertical_marker(self, marker_position=None, mouse_clicked=False):
        """
        The function is called when setting the position of the marker interactively

        If the parameter `marker_position` is `None`, then don't set or change the value.
        Just make the marker visible.
        """

        # Ignore the new value if it is outside the range of selected energies.
        # If 'marker_position' is None, then show the marker at its current location.
        # Totally ignore clicks if 'marker_position' is outside the range (but still
        # display the marker if 'mouse_clicked' is False.
        marker_in_range = True
        if marker_position is not None:
            e_low = self.param_model.param_new["non_fitting_values"]["energy_bound_low"]["value"]
            e_high = self.param_model.param_new["non_fitting_values"]["energy_bound_high"]["value"]
            if e_low <= marker_position <= e_high or not mouse_clicked:
                # If the function was called to display marker (e.g. for existing peak) outside
                #   the selected range, then show it. If button was clicked, then ignore it.
                self.vertical_marker_kev = marker_position
            else:
                marker_in_range = False

        if marker_in_range:
            # Make the marker visible
            self.vertical_marker_is_visible = True

            # Compute peak intensity. The displayed value will change only for user defined peak,
            #   since it is moved to the position of the marker.
            self.compute_manual_peak_intensity()

            # Update the location of the marker and the canvas
            self.plot_vertical_marker()
            self._update_canvas()

            if mouse_clicked:
                try:
                    self.report_marker_state(True)  # This is an externally set callback function
                except Exception:
                    pass

    def hide_plot_vertical_marker(self, mouse_clicked=False):
        """Hide vertical marker"""
        self.vertical_marker_is_visible = False
        self.plot_vertical_marker()
        self._update_canvas()

        if mouse_clicked:
            try:
                self.report_marker_state(False)  # This is an externally set callback function
            except Exception:
                pass

    def plot_selected_energy_range_original(self, *, e_low=None, e_high=None):
        """
        Plot the range of energies selected for processing. The range may be optionally
        provided as arguments. The range values that are not provided, are read from
        globally accessible dictionary of parameters. The values passed as arguments
        are mainly used if the function is called during interactive update of the
        range, when the order of update is undetermined and the parameter dictionary
        may be updated after the function is called.
        """
        # The range of energy selected for analysis
        if e_low is None:
            e_low = self.param_model.param_new["non_fitting_values"]["energy_bound_low"]["value"]
        if e_high is None:
            e_high = self.param_model.param_new["non_fitting_values"]["energy_bound_high"]["value"]

        n_x = 4096  # Set to the maximum possible number of points

        # Generate the values for 'energy' axis
        x_v = (
            self.param_model.param_new["e_offset"]["value"]
            + np.arange(n_x) * self.param_model.param_new["e_linear"]["value"]
            + np.arange(n_x) ** 2 * self.param_model.param_new["e_quadratic"]["value"]
        )

        ss = (x_v < e_high) & (x_v > e_low)
        y_min, y_max = -1e30, 1e30  # Select the max and min values for plotted rectangles

        # Remove the plot if it exists
        if self.plot_energy_barh in self._ax.collections:
            self._ax.collections.remove(self.plot_energy_barh)

        # Create the new plot (based on new parameters if necessary
        self.plot_energy_barh = BrokenBarHCollection.span_where(
            x_v, ymin=y_min, ymax=y_max, where=ss, facecolor="white", edgecolor="yellow", alpha=1
        )
        self._ax.add_collection(self.plot_energy_barh)

    def plot_multi_exp_data(self):
        while len(self.plot_exp_list):
            self.plot_exp_list.pop().remove()

        color_n = get_color_name()

        self.max_v = 1.0
        m = 0
        for (k, v) in self.io_model.data_sets.items():
            if v.selected_for_preview:

                data_arr = np.asarray(v.data)
                # Truncate the array (1D spectrum)
                data_arr = data_arr[0 : self.number_pts_to_show]
                self.max_v = np.max([self.max_v, np.max(data_arr[self.limit_cut : -self.limit_cut])])

                x_v = (
                    self.param_model.param_new["e_offset"]["value"]
                    + np.arange(len(data_arr)) * self.param_model.param_new["e_linear"]["value"]
                    + np.arange(len(data_arr)) ** 2 * self.param_model.param_new["e_quadratic"]["value"]
                )

                (plot_exp_obj,) = self._ax.plot(
                    x_v,
                    data_arr,
                    color=color_n[m],
                    label=v.filename.split(".")[0],
                    linestyle=self.plot_style["experiment"]["linestyle"],
                    marker=self.plot_style["experiment"]["marker"],
                )
                self.plot_exp_list.append(plot_exp_obj)
                m += 1

        self.plot_selected_energy_range_original()

        self._update_ylimit()
        self.log_linear_plot()
        self._update_canvas()

    def plot_emission_line(self):
        """
        Plot emission line and escape peaks associated with given lines.
        The value of self.max_v is needed in this function in order to plot
        the relative height of each emission line.
        """
        while len(self.eline_obj):
            self.eline_obj.pop().remove()

        escape_e = self.escape_e

        if len(self.elist):
            for i in range(len(self.elist)):
                (eline,) = self._ax.plot(
                    [self.elist[i][0], self.elist[i][0]],
                    [0, self.elist[i][1] * self.max_v],
                    color=self.plot_style["emission_line"]["color"],
                    linewidth=self.plot_style["emission_line"]["linewidth"],
                )
                self.eline_obj.append(eline)
                if self.plot_escape_line and self.elist[i][0] > escape_e:
                    (eline,) = self._ax.plot(
                        [self.elist[i][0] - escape_e, self.elist[i][0] - escape_e],
                        [0, self.elist[i][1] * self.max_v],
                        color=self.plot_style["escape"]["color"],
                        linewidth=self.plot_style["emission_line"]["linewidth"],
                    )
                    self.eline_obj.append(eline)

    def _set_eline_select_controls(self, *, element_id=None, data="use_self_data"):
        if element_id is None:
            element_id = self.element_id
        if data == "use_self_data":
            data = self.io_model.data

    def is_line_in_selected_list(self, n_id):
        """
        Checks if the line with ID ``n_id`` is in the list of
        selected element lines.
        Used to enable/disable 'Add Line' and 'Remove Line' buttons.

        Parameters
        ----------

        n_id : Int
            index of the element emission line in the list
            (often equal to ``self.element_id``)

        Returns True if the element line
        is in the list of selected lines. False otherwise.
        """

        ename = self.get_element_line_name_by_id(n_id)

        if ename is None:
            return False

        if self.param_model.EC.is_element_in_list(ename):
            return True
        else:
            return False

    def is_element_line_id_valid(self, n_id):
        """
        Checks if ID (``n_id``) of the element emission line is valid,
        i.e. the name of the line may be obtained by using the ID.

        Parameters
        ----------

        n_id : Int
            index of the element emission line in the list
            (often equal to 'self.element_id')

        Returns True if the element line is valid
        """

        # There may be a more efficient way to check 'n_id',
        #   but we want to use the same function as we use
        #   to retrive the line name
        ename = self.get_element_line_name_by_id(n_id)

        if ename is None:
            return False
        else:
            return True

    def get_element_line_name_by_id(self, n_id):
        """
        Retrieves the name of the element emission line from its ID
        (the number in the list). The lines are numbered starting with 1.
        If the ID is invalid, the function returns None.

        Parameters
        ----------

        n_id : int
            index of the element emission line in the list
            (often equal to 'self.element_id')

        Returns the line name (str). If the name can not be retrieved, then
        the function returns None.
        """
        if n_id < 1:
            # Elements are numbered starting with 1. Element #0 does not exist.
            #   (Element #0 means that no element is selected)
            return None

        # This is the fixed list of element emission line names.
        #   The element with ID==1 is found in total_list[0]
        total_list = self.param_model.get_user_peak_list()
        try:
            ename = total_list[n_id - 1]
        except Exception:
            ename = None
        return ename

    def _vertical_marker_set_inside_range(self, *, e_low=None, e_high=None):
        """
        Don't move the marker if it is inside range. If it is outside range,
        then set the marker to the center of the range
        """
        # The range of energy selected for analysis
        if e_low is None:
            e_low = self.param_model.param_new["non_fitting_values"]["energy_bound_low"]["value"]
        if e_high is None:
            e_high = self.param_model.param_new["non_fitting_values"]["energy_bound_high"]["value"]

        # By default, place the marker in the middle of the range if its original position
        #   is outside the range
        if (self.vertical_marker_kev > e_high) or (self.vertical_marker_kev < e_low):
            self.vertical_marker_kev = (e_low + e_high) / 2.0

    def _fill_elist(self):

        _elist = []

        incident_energy = self.incident_energy
        k_len = len(K_TRANSITIONS)
        l_len = len(L_TRANSITIONS)
        m_len = len(M_TRANSITIONS)

        ename = self.get_element_line_name_by_id(self.element_id)

        if ename is not None:

            _elist = []
            if ename.lower().startswith("userpeak"):
                # Make sure that the marker is in the selected range of energies
                self._vertical_marker_set_inside_range()
                # The tuple structure: (center_energy, ratio)
                _elist.append((self.vertical_marker_kev, 1.0))

            elif "_K" in ename:
                e = Element(ename[:-2])
                if e.cs(incident_energy)["ka1"] != 0:
                    for i in range(k_len):
                        _elist.append(
                            (
                                e.emission_line.all[i][1],
                                e.cs(incident_energy).all[i][1] / e.cs(incident_energy).all[0][1],
                            )
                        )

            elif "_L" in ename:
                e = Element(ename[:-2])
                if e.cs(incident_energy)["la1"] != 0:
                    for i in range(k_len, k_len + l_len):
                        _elist.append(
                            (
                                e.emission_line.all[i][1],
                                e.cs(incident_energy).all[i][1] / e.cs(incident_energy).all[k_len][1],
                            )
                        )

            else:
                e = Element(ename[:-2])
                if e.cs(incident_energy)["ma1"] != 0:
                    for i in range(k_len + l_len, k_len + l_len + m_len):
                        _elist.append(
                            (
                                e.emission_line.all[i][1],
                                e.cs(incident_energy).all[i][1] / e.cs(incident_energy).all[k_len + l_len][1],
                            )
                        )

        return _elist

    def _get_pileup_lines(self, eline):
        """
        Returns the energy (center) of pileup peak. And the energies of two components.

        Parameters
        ----------
        eline: str
            Name of the pileup peak, e.g. V_Ka1-Co_Ka1

        Returns
        -------
        list(float)
            Energy in keV of pileup peak and two components
        """
        try:
            element_line1, element_line2 = eline.split("-")
            e1_cen = get_eline_parameters(element_line1, self.incident_energy)["energy"]
            e2_cen = get_eline_parameters(element_line2, self.incident_energy)["energy"]
            en = [e1_cen + e2_cen, e1_cen, e2_cen]
        except Exception:
            en = []
        return en

    def _fill_elist_pileup(self, eline=None):
        if eline is None:
            eline = self.param_model.e_name

        elist = []
        energies = self._get_pileup_lines(eline)
        if energies:
            elist = list(zip(energies, [1, 0.2, 0.2]))

        return elist

    def _fill_elist_userpeak(self):
        """
        Fill the list of 'emission lines' for user defined peak. There is only ONE
        'emission line', with position determined by the location of the marker.
        If the marker is not currently visible, then don't put any emission lines in the list.
        The list is used during adding user-defined peaks.
        """
        elist = []
        energy, marker_visible = self.get_suggested_new_manual_peak_energy()
        if marker_visible:
            elist.append((energy, 1))
        return elist

    def _reset_eline_plot(self):
        while len(self.eline_obj):
            self.eline_obj.pop().remove()
        self.elist = []
        self._fig.canvas.draw()

    @observe("element_id")
    def set_element(self, change):

        self._set_eline_select_controls(element_id=change["value"])
        self.compute_manual_peak_intensity(n_id=change["value"])

        if change["value"] == 0:
            self._reset_eline_plot()
            return

        self.plot_current_eline()

    def plot_current_eline(self, eline=None):
        """
        Plots emission lines for the selected peak based on 'self.element_id` and provided `eline`.
        """
        if eline is None:
            eline = self.param_model.e_name

        incident_energy = self.incident_energy

        # Name of the emission line (if emission line is selected)
        ename = self.get_element_line_name_by_id(self.element_id)
        # Check if pileup peak is selected
        is_pileup = self.param_model.get_eline_name_category(eline) == "pileup"

        if (ename is not None) or is_pileup:

            logger.debug(
                "Plot emission line for element: "
                "{} with incident energy {}".format(self.element_id, incident_energy)
            )

            if ename is not None:
                self.elist = self._fill_elist()
            elif is_pileup:
                self.elist = self._fill_elist_pileup(eline)
            else:
                self.elist = []  # Just in case
            self.plot_emission_line()
            self._update_canvas()

            # Do it the second time, since the 'self.elist' has changed
            self.compute_manual_peak_intensity(n_id=self.element_id)

        else:
            self._reset_eline_plot()
            logger.debug(f"Selected emission line with ID #{self.element_id} is not in the list.")

    @observe("det_materials")
    def _update_det_materials(self, change):
        if change["value"] == 0:
            self.escape_e = 1.73998
        else:
            self.escape_e = 9.88640

    def change_escape_peak_settings(self, plot_escape_line, det_material):
        self.plot_escape_line = plot_escape_line
        self.det_materials = det_material
        # Now update the displayed emission line
        self.plot_emission_line()
        self._update_canvas()

    def plot_roi_bound(self):
        """
        Plot roi with low, high and ceter value.
        """
        for k, v in self.roi_plot_dict.items():
            for data in v:
                data.remove()
        self.roi_plot_dict.clear()

        if len(self.roi_dict):
            # self._ax.hold(True)
            for k, v in self.roi_dict.items():
                temp_list = []
                for linev in np.array([v.left_val, v.line_val, v.right_val]) / 1000.0:
                    (lineplot,) = self._ax.plot(
                        [linev, linev],
                        [0, 1 * self.max_v],
                        color=self.plot_style["roi_line"]["color"],
                        linewidth=self.plot_style["roi_line"]["linewidth"],
                    )
                    if v.show_plot:
                        lineplot.set_visible(True)
                    else:
                        lineplot.set_visible(False)
                    temp_list.append(lineplot)
                self.roi_plot_dict.update({k: temp_list})

        self._update_canvas()

    @observe("roi_dict")
    def show_roi_bound(self, change):
        logger.debug("roi dict changed {}".format(change["value"]))
        self.plot_roi_bound()

        if len(self.roi_dict):
            for k, v in self.roi_dict.items():
                if v.show_plot:
                    for ln in self.roi_plot_dict[k]:
                        ln.set_visible(True)
                else:
                    for ln in self.roi_plot_dict[k]:
                        ln.set_visible(False)
        self._update_canvas()

    def get_suggested_new_manual_peak_energy(self):
        """
        Returns energy pointed by the vertical marker in keV and the status of the marker.

        Returns
        -------
        float
            Energy of the manual peak center in keV. The energy is determined
            by vertical marker on the screen.
        bool
            True if the vertical marker is visible, otherwise False.
        """
        energy = self.vertical_marker_kev
        marker_visible = self.vertical_marker_is_visible
        return energy, marker_visible

    def _compute_intensity(self, elist):
        # Some default value
        intensity = 1000.0

        if (
            self.io_model.data is not None
            and self.param_model.param_new is not None
            and self.param_model.prefit_x is not None
            and self.param_model.total_y is not None
            and len(self.io_model.data) > 1
            and len(self.param_model.prefit_x) > 1
        ):

            # Range of energies in fitting results
            e_fit_min = self.param_model.prefit_x[0]
            e_fit_max = self.param_model.prefit_x[-1]
            de_fit = (e_fit_max - e_fit_min) / (len(self.param_model.prefit_x) - 1)

            e_raw_min = self.param_model.param_new["e_offset"]["value"]
            e_raw_max = (
                self.param_model.param_new["e_offset"]["value"]
                + (len(self.io_model.data) - 1) * self.param_model.param_new["e_linear"]["value"]
                + (len(self.io_model.data) - 1) ** 2 * self.param_model.param_new["e_quadratic"]["value"]
            )

            de_raw = (e_raw_max - e_raw_min) / (len(self.io_model.data) - 1)

            # Note: the above algorithm for finding 'de_raw' is far from perfect but will
            #    work for now. As a result 'de_fit' and
            #    'de_raw' == sself.param_model.param_new['e_linear']['value'].
            #    So the quadratic coefficent is ignored. This is OK, since currently
            #    quadratic coefficient is always ZERO. When the program is rewritten,
            #    the complete algorithm should be revised.

            # Find the line with maximum energy. It must come first in the list,
            #    but let's check just to make sure
            max_line_energy, max_line_intensity = 0, 0
            if elist:
                for e, i in elist:
                    # e - line peak energy
                    # i - peak intensity relative to maximum peak
                    if e >= e_fit_min and e <= e_fit_max and e > e_raw_min and e < e_raw_max:
                        if max_line_intensity < i:
                            max_line_energy, max_line_intensity = e, i

            # Find the index of peak maximum in the 'fitted' data array
            n = (max_line_energy - e_fit_min) / de_fit
            n = np.clip(n, 0, len(self.param_model.total_y) - 1)
            n_fit = int(round(n))
            # Find the index of peak maximum in the 'raw' data array
            n = (max_line_energy - e_raw_min) / de_raw
            n = np.clip(n, 0, len(self.io_model.data) - 1)
            n_raw = int(round(n))
            # Intensity of the fitted data at the peak
            in_fit = self.param_model.total_y[n_fit]
            # Intensity of the raw data at the peak
            in_raw = self.io_model.data[n_raw]
            # The estimated peak intensity is the difference:
            intensity = in_raw - in_fit

            # The following step is questionable. We assign some reasonably small number.
            #   The desired value can always be manually entered
            if intensity < 0.0:
                intensity = abs(in_raw / 100)

        return intensity

    def compute_manual_peak_intensity(self, n_id=None):

        if n_id is None:
            n_id = self.element_id

        # Check if the emission line is in the list of supported emission lines (e.g. Ca_K)
        if not self.is_element_line_id_valid(n_id):
            # This is not a supported emission line (n_id==0)
            # This means we are probably dealing with user defined peak.
            if self.is_line_in_selected_list(n_id):
                # Display intensity if the peak is in the list.
                name = self.get_element_line_name_by_id(n_id)
                intensity = self.param_model.EC.element_dict[name].maxv
            else:
                elist = self._fill_elist_userpeak()
                intensity = self._compute_intensity(elist)
        else:
            if self.is_line_in_selected_list(n_id):
                # Display intensity if the peak is in the list.
                name = self.get_element_line_name_by_id(n_id)
                intensity = self.param_model.EC.element_dict[name].maxv
            else:
                # This is a new peak
                elist = self._fill_elist()
                intensity = self._compute_intensity(elist)

        # Round the intensity for nicer printing
        self.param_model.add_element_intensity = round(intensity, 2)

    def plot_fit(self, fit_x, fit_y, fit_all, residual=None):
        """
        Parameters
        ----------
        fit_x : array
            energy axis
        fit_y : array
            fitted spectrum
        fit_all : dict
            dict of individual line
        residual : array
            residual between fit and exp
        """
        if fit_x is None or fit_y is None:
            return

        while len(self.plot_fit_obj):
            self.plot_fit_obj.pop().remove()

        (ln,) = self._ax.plot(
            fit_x,
            fit_y,
            color=self.plot_style["fit"]["color"],
            label=self.plot_style["fit"]["label"],
            linewidth=self.plot_style["fit"]["linewidth"],
        )
        self.plot_fit_obj.append(ln)

        if residual is not None:
            # shiftv = 1.5  # move residual down by some amount
            (ln,) = self._ax.plot(
                fit_x,
                residual - 0.15 * self.max_v,  # shiftv*(np.max(np.abs(self.residual))),
                label=self.plot_style["residual"]["label"],
                color=self.plot_style["residual"]["color"],
            )
            self.plot_fit_obj.append(ln)

        k_num = 0
        l_num = 0
        m_num = 0
        p_num = 0
        for k, v in fit_all.items():
            if k == "background":
                (ln,) = self._ax.plot(
                    fit_x,
                    v,
                    color=self.plot_style["background"]["color"],
                    # marker=self.plot_style['background']['marker'],
                    # markersize=self.plot_style['background']['markersize'],
                    label=self.plot_style["background"]["label"],
                )
                self.plot_fit_obj.append(ln)
            elif k == "compton":
                (ln,) = self._ax.plot(
                    fit_x,
                    v,
                    color=self.plot_style["compton"]["color"],
                    linewidth=self.plot_style["compton"]["linewidth"],
                    label=self.plot_style["compton"]["label"],
                )
                self.plot_fit_obj.append(ln)
            elif k == "elastic":
                (ln,) = self._ax.plot(
                    fit_x, v, color=self.plot_style["elastic"]["color"], label=self.plot_style["elastic"]["label"]
                )
                self.plot_fit_obj.append(ln)
            elif k == "escape":
                (ln,) = self._ax.plot(
                    fit_x, v, color=self.plot_style["escape"]["color"], label=self.plot_style["escape"]["label"]
                )
                self.plot_fit_obj.append(ln)

            elif "user" in k.lower():
                (ln,) = self._ax.plot(
                    fit_x,
                    v,
                    color=self.plot_style["userpeak"]["color"],
                    label=self.plot_style["userpeak"]["label"],
                )
                self.plot_fit_obj.append(ln)

            elif "-" in k:  # Si_K-Si_K
                if p_num == 0:
                    (ln,) = self._ax.plot(
                        fit_x,
                        v,
                        color=self.plot_style["pileup"]["color"],
                        label=self.plot_style["pileup"]["label"],
                    )
                else:
                    (ln,) = self._ax.plot(fit_x, v, color=self.plot_style["pileup"]["color"], label="_nolegend_")
                self.plot_fit_obj.append(ln)
                p_num += 1

            elif ("_K" in k.upper()) and (len(k) <= 4):
                if k_num == 0:
                    (ln,) = self._ax.plot(
                        fit_x,
                        v,
                        color=self.plot_style["k_line"]["color"],
                        label=self.plot_style["k_line"]["label"],
                    )
                else:
                    (ln,) = self._ax.plot(fit_x, v, color=self.plot_style["k_line"]["color"], label="_nolegend_")
                self.plot_fit_obj.append(ln)
                k_num += 1

            elif ("_L" in k.upper()) and (len(k) <= 4):
                if l_num == 0:
                    (ln,) = self._ax.plot(
                        fit_x,
                        v,
                        color=self.plot_style["l_line"]["color"],
                        label=self.plot_style["l_line"]["label"],
                    )
                else:
                    (ln,) = self._ax.plot(fit_x, v, color=self.plot_style["l_line"]["color"], label="_nolegend_")
                self.plot_fit_obj.append(ln)
                l_num += 1

            elif ("_M" in k.upper()) and (len(k) <= 4):
                if m_num == 0:
                    (ln,) = self._ax.plot(
                        fit_x,
                        v,
                        color=self.plot_style["m_line"]["color"],
                        label=self.plot_style["m_line"]["label"],
                    )
                else:
                    (ln,) = self._ax.plot(fit_x, v, color=self.plot_style["m_line"]["color"], label="_nolegend_")
                self.plot_fit_obj.append(ln)
                m_num += 1

            else:
                pass

        # self._update_canvas()

    def canvas_onpress(self, event):
        """Callback, mouse button pressed"""
        if self.t_bar.mode == "":
            if event.inaxes == self._ax:
                if event.button == 1:
                    xd = event.xdata
                    self.set_plot_vertical_marker(marker_position=xd, mouse_clicked=True)
                else:
                    self.hide_plot_vertical_marker(mouse_clicked=True)

    # ===========================================================
    #         Functions for plotting spectrum preview

    def selected_range_indices(self, *, e_low=None, e_high=None, n_indexes=None, margin=2.0):
        """
        The function computes the range of indices based on the selected energy range
        and parameters for the energy axis.

        Parameters
        ----------
        e_low, e_high: float or None
            Energy values (in keV) that set the selected range
        n_indexes: int
            Total number of indexes in the energy array (typically 4096)
        margin: float
            The displayed energy range is extended by the value of `margin` in both directions.

        Returns
        -------
        n_low, n_high: int
            The range of indices of the energy array (n_low..n_high-1) that cover the selected energy range
        """
        # The range of energy selected for analysis
        if e_low is None:
            e_low = self.param_model.param_new["non_fitting_values"]["energy_bound_low"]["value"]
        if e_high is None:
            e_high = self.param_model.param_new["non_fitting_values"]["energy_bound_high"]["value"]
        # Protection for the case if e_high < e_low
        e_high = e_high if e_high > e_low else e_low
        # Extend the range (by the value of 'margin')
        e_low, e_high = e_low - margin, e_high + margin

        # The following calculations ignore quadratic term, which is expected to be small
        c0 = self.param_model.param_new["e_offset"]["value"]
        c1 = self.param_model.param_new["e_linear"]["value"]
        # If more precision if needed, then implement more complicated algorithm using
        #   the quadratic term: c2 = self.param_model.param_new['e_quadratic']['value']

        n_low = int(np.clip(int((e_low - c0) / c1), a_min=0, a_max=n_indexes - 1))
        n_high = int(np.clip(int((e_high - c0) / c1) + 1, a_min=1, a_max=n_indexes))

        return n_low, n_high

    def _datasets_max_size(self, *, only_displayed=True):
        """
        Return maximum size of the longest available dataset. The datasets that contain
        no data are ignored.

        Parameters
        ----------
        only_displayed: bool
            Limit search to the datasets that are going to be displayed
        """
        max_size = 0
        for dset in self.io_model.data_sets.values():
            if not only_displayed or dset.selected_for_preview:
                # Raw data shape: (n_rows, n_columns, n_energy_bins)
                max_size = max(max_size, dset.get_raw_data_shape()[2])
        return max_size

    def plot_selected_energy_range(self, *, axes, barh_existing, e_low=None, e_high=None, n_points=4096):
        """
        Plot the range of energies selected for processing. The range may be optionally
        provided as arguments. The range values that are not provided, are read from
        globally accessible dictionary of parameters. The values passed as arguments
        are mainly used if the function is called during interactive update of the
        range, when the order of update is undetermined and the parameter dictionary
        may be updated after the function is called.
        """
        # The range of energy selected for analysis
        if e_low is None:
            e_low = self.param_model.param_new["non_fitting_values"]["energy_bound_low"]["value"]
        if e_high is None:
            e_high = self.param_model.param_new["non_fitting_values"]["energy_bound_high"]["value"]

        # Model coefficients for the energy axis
        c0 = self.param_model.param_new["e_offset"]["value"]
        c1 = self.param_model.param_new["e_linear"]["value"]
        c2 = self.param_model.param_new["e_quadratic"]["value"]

        # Generate the values for 'energy' axis
        x_v = c0 + np.arange(n_points) * c1 + np.arange(n_points) ** 2 * c2
        ss = (x_v < e_high + c1) & (x_v > e_low - c1)

        # Trim both arrays to minimize the number of points
        x_v = x_v[ss]
        ss = ss[ss]
        ss[0] = False
        ss[-1] = False

        # Negative values will work for semilog plot as well
        y_min, y_max = -1e30, 1e30  # Select the max and min values for plotted rectangles

        # Remove the plot if it exists
        if barh_existing in axes.collections:
            axes.collections.remove(barh_existing)

        # Create the new plot (based on new parameters if necessary
        barh_new = BrokenBarHCollection.span_where(
            x_v, ymin=y_min, ymax=y_max, where=ss, facecolor="white", edgecolor="yellow", alpha=1
        )
        axes.add_collection(barh_new)

        return barh_new

    def prepare_preview_spectrum_plot(self):

        if self._ax_preview:
            self._ax_preview.clear()
        else:
            self._ax_preview = self._fig_preview.add_subplot(111)
        self._ax_preview.set_facecolor("lightgrey")
        self._ax_preview.grid(which="both")

        self._fig_preview.set_visible(False)

    def _show_preview_spectrum_plot(self):

        # Completely redraw the plot each time the function is called
        self.prepare_preview_spectrum_plot()

        # Remove all lines from the plot
        while len(self._lines_preview):
            self._lines_preview.pop().remove()

        # The list of color names
        color_names = get_color_name()

        e_range = self.energy_range_preview
        e_range_supported = (EnergyRangePresets.SELECTED_RANGE, EnergyRangePresets.FULL_SPECTRUM)
        if e_range not in e_range_supported:
            logger.error(
                f"Spectrum preview: Unknown option for the energy range: {e_range}\n"
                "Please report the error to the development team."
            )
            # This is not a critical error, so we still can proceed
            e_range = EnergyRangePresets.FULL_SPECTRUM

        p_type = self.plot_type_preview
        p_type_supported = (PlotTypes.LINLOG, PlotTypes.LINEAR)
        if p_type not in p_type_supported:
            logger.error(
                f"Spectrum preview: Unknown option for the plot type: {p_type}\n"
                "Please report the error to the development team."
            )
            p_type = PlotTypes.LINEAR

        # Maximum number of points in the displayed dataset
        n_dset_points = self._datasets_max_size()

        if e_range == EnergyRangePresets.SELECTED_RANGE:
            n_range_low, n_range_high = self.selected_range_indices(n_indexes=n_dset_points)
        else:
            n_range_low, n_range_high = 0, n_dset_points

        # All available datasets, we will print only the selected datasets
        dset_names = list(self.io_model.data_sets.keys())

        if p_type == PlotTypes.LINLOG:
            top_margin_coef = 2.0
            # Minimum for semilog plots may need to be computed, but 1.0 is good
            self.min_v_preview = 1.0
            self._ax_preview.set_yscale("log")
        else:
            top_margin_coef = 1.05
            self.min_v_preview = 0.0  # Minimum will always be 0 for linear plots

        self.max_v_preview = 1.0
        self.min_e_preview = 1000.0  # Start with some large number
        self.max_e_preview = 0.1  # Start with some small number
        for n_line, dset_name in enumerate(dset_names):
            dset = self.io_model.data_sets[dset_name]

            # Select color (even if the dataset is not displayed). This is done in order
            #   to ensure that each dataset is assigned the unique color.
            color = color_names[n_line % len(color_names)]

            if dset.selected_for_preview:

                data_arr = np.asarray(dset.get_total_spectrum())
                if data_arr is None:  # Just a precaution, it shouldn't happen
                    logger.error("Spectrum review: attempting to print empty dataset.")
                    continue

                # The assumption is that some datasets may have different length (which is
                #   currently not the case). So we have to take it into account when using
                #   maximum dataset length. This is essentially a safety precaution.
                n_low = int(np.clip(n_range_low, a_min=0, a_max=data_arr.size - 1))
                n_high = int(np.clip(n_range_high, a_min=1, a_max=data_arr.size))

                # From now on we work with the trimmed data array
                x_v = (
                    self.param_model.param_new["e_offset"]["value"]
                    + np.arange(n_low, n_high) * self.param_model.param_new["e_linear"]["value"]
                    + np.arange(n_low, n_high) ** 2 * self.param_model.param_new["e_quadratic"]["value"]
                )

                data_arr = data_arr[n_low:n_high]

                self.max_v_preview = np.max(
                    [self.max_v_preview, np.max(data_arr[self.limit_cut : -self.limit_cut])]
                )
                self.max_e_preview = np.max([self.max_e_preview, x_v[-1]])
                self.min_e_preview = np.min([self.min_e_preview, x_v[0]])

                (line,) = self._ax_preview.plot(
                    x_v,
                    data_arr,
                    color=color,
                    label=dset.filename.split(".")[0],
                    linestyle=self.plot_style["experiment"]["linestyle"],
                    marker=self.plot_style["experiment"]["marker"],
                )

                self._lines_preview.append(line)

        self._ax_preview.set_xlim(self.min_e_preview, self.max_e_preview)
        self._ax_preview.set_ylim(self.min_v_preview, self.max_v_preview * top_margin_coef)
        self._ax_preview.legend()
        self._ax_preview.set_xlabel("Energy (keV)")
        self._ax_preview.set_ylabel("Total Spectrum (Counts)")
        self._fig_preview.set_visible(True)

        # Reset navigation toolbar (specifically clear ZOOM history, since it becomes invalid
        #   when the new data is loaded, i.e. zooming out may not show the whole plot)
        tb = self._fig_preview.canvas.toolbar
        tb.update()

        self._bahr_preview = self.plot_selected_energy_range(
            axes=self._ax_preview, barh_existing=self._bahr_preview
        )

    def _hide_preview_spectrum_plot(
        self,
    ):
        self._fig_preview.set_visible(False)

    def update_preview_spectrum_plot(self, *, hide=False):
        """
        Update spectrum preview plot based on available/selected dataset and `hide` flag.

        Parameters
        ----------
        hide: bool
            `False` - plot data if datasets are available and at least one dataset is selected,
            otherwise hide the plot, `True` - hide the plot in any case
        """
        # Find out if any data is selected
        show_plot = False
        if self.io_model.data_sets:
            show_plot = any([_.selected_for_preview for _ in self.io_model.data_sets.values()])
        logger.debug(f"LinePlotModel.update_preview_spectrum_plot(): show_plot={show_plot} hide={hide}")
        if show_plot and not hide:
            logger.debug("LinePlotModel.update_preview_spectrum_plot(): plotting existing datasets")
            self._show_preview_spectrum_plot()
        else:
            logger.debug("LinePlotModel.update_preview_spectrum_plot(): hiding plots")
            self._hide_preview_spectrum_plot()
        self._fig_preview.canvas.draw()

    # ===========================================================================================
    #   Plotting the preview of Total Count Maps

    def clear_map_preview_range(self):
        self.set_map_preview_range(low=-1, high=-1)

    def set_map_preview_range(self, *, low, high):
        self.map_preview_range_low = low
        self.map_preview_range_high = high

    def get_selected_datasets(self):
        """Returns the datasets selected for preview"""
        return {k: v for (k, v) in self.io_model.data_sets.items() if v.selected_for_preview}

    def _compute_map_preview_range(self, img_dict, key_list):
        range_min, range_max = None, None
        for key in key_list:
            data = img_dict[key]
            v_min, v_max = np.min(data), np.max(data)
            if range_min is None or range_max is None:
                range_min, range_max = v_min, v_max
            else:
                range_min, range_max = min(range_min, v_min), max(range_max, v_max)
        return range_min, range_max

    def _show_total_count_map_preview(self):
        self._fig_maps.set_visible(True)
        self._fig_maps.clf()
        selected_dsets = self.get_selected_datasets()
        data_for_plotting = {k: v.get_total_count() for (k, v) in selected_dsets.items()}

        # Check if positions data is available. Positions data may be unavailable
        # (not recorded in HDF5 file) if experiment is has not been completed.
        # While the data from the completed part of experiment may still be used,
        # plotting vs. x-y or scatter plot may not be displayed.
        positions_data_available = False
        if "positions" in self.io_model.img_dict.keys():
            data_for_plotting["positions"] = self.io_model.img_dict["positions"]
            positions_data_available = True

        # Create local copies of self.pixel_or_pos, self.scatter_show and self.grid_interpolate
        pixel_or_pos_local = self.map_axes_units_preview
        scatter_show_local = self.map_scatter_plot

        # Disable plotting vs x-y coordinates if 'positions' data is not available
        if not positions_data_available:
            if pixel_or_pos_local:
                pixel_or_pos_local = MapAxesUnits.PIXELS  # Switch to plotting vs. pixel number
                logger.error("'Positions' data is not available. Plotting vs. x-y coordinates is disabled")
            if scatter_show_local:
                scatter_show_local = False  # Switch to plotting vs. pixel number
                logger.error("'Positions' data is not available. Scatter plot is disabled.")

        # low_lim = 1e-4  # define the low limit for log image
        plot_interp = "Nearest"

        grey_use = self.map_preview_color_scheme

        ncol = int(np.ceil(np.sqrt(len(selected_dsets))))
        try:
            nrow = int(np.ceil(len(selected_dsets) / float(ncol)))
        except ZeroDivisionError:
            ncol = 1
            nrow = 1

        a_pad_v = 0.8
        a_pad_h = 0.5

        n_displayed_axes = ncol * nrow  # Total number of axes in the grid

        grid = ImageGrid(
            self._fig_maps,
            111,
            nrows_ncols=(nrow, ncol),
            axes_pad=(a_pad_v, a_pad_h),
            cbar_location="right",
            cbar_mode="each",
            cbar_size="7%",
            cbar_pad="2%",
            share_all=True,
        )

        def _compute_equal_axes_ranges(x_min, x_max, y_min, y_max):
            """
            Compute ranges for x- and y- axes of the plot. Make sure that the ranges for x- and y-axes are
            always equal and fit the maximum of the ranges for x and y values:
                  max(abs(x_max-x_min), abs(y_max-y_min))
            The ranges are set so that the data is always centered in the middle of the ranges

            Parameters
            ----------

            x_min, x_max, y_min, y_max : float
                lower and upper boundaries of the x and y values

            Returns
            -------

            x_axis_min, x_axis_max, y_axis_min, y_axis_max : float
                lower and upper boundaries of the x- and y-axes ranges
            """

            x_axis_min, x_axis_max, y_axis_min, y_axis_max = x_min, x_max, y_min, y_max
            x_range, y_range = abs(x_max - x_min), abs(y_max - y_min)
            if x_range > y_range:
                y_center = (y_max + y_min) / 2
                y_axis_max = y_center + x_range / 2
                y_axis_min = y_center - x_range / 2
            else:
                x_center = (x_max + x_min) / 2
                x_axis_max = x_center + y_range / 2
                x_axis_min = x_center - y_range / 2

            return x_axis_min, x_axis_max, y_axis_min, y_axis_max

        def _adjust_data_range_using_min_ratio(c_min, c_max, c_axis_range, *, min_ratio=0.01):
            """
            Adjust the range for plotted data along one axis (x or y). The adjusted range is
            applied to the 'extend' attribute of imshow(). The adjusted range is always greater
            than 'axis_range * min_ratio'. Such transformation has no physical meaning
            and performed for aesthetic reasons: stretching the image presentation of
            a scan with only a few lines (1-3) greatly improves visibility of data.

            Parameters
            ----------

            c_min, c_max : float
                boundaries of the data range (along x or y axis)
            c_axis_range : float
                range presented along the same axis

            Returns
            -------

            cmin, c_max : float
                adjusted boundaries of the data range
            """
            c_range = c_max - c_min
            if c_range < c_axis_range * min_ratio:
                c_center = (c_max + c_min) / 2
                c_new_range = c_axis_range * min_ratio
                c_min = c_center - c_new_range / 2
                c_max = c_center + c_new_range / 2
            return c_min, c_max

        # Hide the axes that are unused (they are unsightly)
        for i in range(len(selected_dsets), n_displayed_axes):
            grid[i].set_visible(False)
            grid.cbar_axes[i].set_visible(False)

        for i, (k, v) in enumerate(selected_dsets.items()):

            data_arr = data_for_plotting[k]

            if pixel_or_pos_local == MapAxesUnits.POSITIONS or scatter_show_local:

                # xd_min, xd_max, yd_min, yd_max = min(self.x_pos), max(self.x_pos),
                #     min(self.y_pos), max(self.y_pos)
                x_pos_2D = data_for_plotting["positions"]["x_pos"]
                y_pos_2D = data_for_plotting["positions"]["y_pos"]
                xd_min, xd_max, yd_min, yd_max = x_pos_2D.min(), x_pos_2D.max(), y_pos_2D.min(), y_pos_2D.max()
                xd_axis_min, xd_axis_max, yd_axis_min, yd_axis_max = _compute_equal_axes_ranges(
                    xd_min, xd_max, yd_min, yd_max
                )

                xd_min, xd_max = _adjust_data_range_using_min_ratio(xd_min, xd_max, xd_axis_max - xd_axis_min)
                yd_min, yd_max = _adjust_data_range_using_min_ratio(yd_min, yd_max, yd_axis_max - yd_axis_min)

                # Adjust the direction of each axis depending on the direction in which encoder values changed
                #   during the experiment. Data is plotted starting from the upper-right corner of the plot
                if x_pos_2D[0, 0] > x_pos_2D[0, -1]:
                    xd_min, xd_max, xd_axis_min, xd_axis_max = xd_max, xd_min, xd_axis_max, xd_axis_min
                if y_pos_2D[0, 0] > y_pos_2D[-1, 0]:
                    yd_min, yd_max, yd_axis_min, yd_axis_max = yd_max, yd_min, yd_axis_max, yd_axis_min

            else:

                yd, xd = data_arr.shape

                xd_min, xd_max, yd_min, yd_max = 0, xd, 0, yd
                if (yd <= math.floor(xd / 100)) and (xd >= 200):
                    yd_min, yd_max = -math.floor(xd / 200), math.ceil(xd / 200)
                if (xd <= math.floor(yd / 100)) and (yd >= 200):
                    xd_min, xd_max = -math.floor(yd / 200), math.ceil(yd / 200)

                xd_axis_min, xd_axis_max, yd_axis_min, yd_axis_max = _compute_equal_axes_ranges(
                    xd_min, xd_max, yd_min, yd_max
                )

            # Compute range for data values
            low_limit = self.map_preview_range_low
            high_limit = self.map_preview_range_high
            # If limit is not set, then compute the limit based on the selected datasets.
            # It is assumed that at least one dataset is selected.
            if low_limit == -1 and high_limit == -1:
                low_limit, high_limit = self._compute_map_preview_range(data_for_plotting, selected_dsets.keys())
                if low_limit is None or high_limit is None:
                    low_limit, high_limit = 0
            # Set some minimum range for the colorbar (otherwise it will have white fill)
            if math.isclose(low_limit, high_limit, abs_tol=2e-20):
                if abs(low_limit) < 1e-20:  # The value is zero
                    dv = 1e-20
                else:
                    dv = math.fabs(low_limit * 0.01)
                high_limit += dv
                low_limit -= dv

            if self.map_type_preview == MapTypes.LINEAR:
                if not scatter_show_local:
                    im = grid[i].imshow(
                        data_arr,
                        cmap=grey_use,
                        interpolation=plot_interp,
                        extent=(xd_min, xd_max, yd_max, yd_min),
                        origin="upper",
                        clim=(low_limit, high_limit),
                    )
                    grid[i].set_ylim(yd_axis_max, yd_axis_min)
                else:
                    xx = self.io_model.img_dict["positions"]["x_pos"]
                    yy = self.io_model.img_dict["positions"]["y_pos"]

                    # The following condition prevents crash if different file is loaded while
                    #    the scatter plot is open (PyXRF specific issue)
                    if data_arr.shape == xx.shape and data_arr.shape == yy.shape:
                        im = grid[i].scatter(
                            xx,
                            yy,
                            c=data_arr,
                            marker="s",
                            s=500,
                            alpha=1.0,  # Originally: alpha=0.8
                            cmap=grey_use,
                            vmin=low_limit,
                            vmax=high_limit,
                            linewidths=1,
                            linewidth=0,
                        )
                        grid[i].set_ylim(yd_axis_max, yd_axis_min)

                grid[i].set_xlim(xd_axis_min, xd_axis_max)

                grid_title = k
                # Display only the channel name (e.g. 'sum', 'det1' etc.)
                grid_title = grid_title.split("_")[-1]
                grid[i].text(0, 1.01, grid_title, ha="left", va="bottom", transform=grid[i].axes.transAxes)

                grid.cbar_axes[i].colorbar(im)

                im.colorbar.formatter = im.colorbar.ax.yaxis.get_major_formatter()
                # im.colorbar.ax.get_xaxis().set_ticks([])
                # im.colorbar.ax.get_xaxis().set_ticks([], minor=True)
                grid.cbar_axes[i].ticklabel_format(style="sci", scilimits=(-3, 4), axis="both")

            else:

                # maxz = np.max(data_arr)
                # # Set some reasonable minimum range for the colorbar
                # #   Zeros or negative numbers will be shown in white
                # if maxz <= 1e-30:
                #     maxz = 1

                if not scatter_show_local:
                    im = grid[i].imshow(
                        data_arr,
                        # norm=LogNorm(vmin=low_lim*maxz,
                        #              vmax=maxz, clip=True),
                        norm=LogNorm(vmin=low_limit, vmax=high_limit, clip=True),
                        cmap=grey_use,
                        interpolation=plot_interp,
                        extent=(xd_min, xd_max, yd_max, yd_min),
                        origin="upper",
                        # clim=(low_lim*maxz, maxz))
                        clim=(low_limit, high_limit),
                    )

                    grid[i].set_ylim(yd_axis_max, yd_axis_min)
                else:
                    im = grid[i].scatter(
                        self.io_model.img_dict["positions"]["x_pos"],
                        self.io_model.img_dict["positions"]["y_pos"],
                        # norm=LogNorm(vmin=low_lim*maxz,
                        #              vmax=maxz, clip=True),
                        norm=LogNorm(vmin=low_limit, vmax=high_limit, clip=True),
                        c=data_arr,
                        marker="s",
                        s=500,
                        alpha=1.0,  # Originally: alpha=0.8
                        cmap=grey_use,
                        linewidths=1,
                        linewidth=0,
                    )
                    grid[i].set_ylim(yd_axis_min, yd_axis_max)

                grid[i].set_xlim(xd_axis_min, xd_axis_max)

                grid_title = k
                # Display only the channel name (e.g. 'sum', 'det1' etc.)
                grid_title = grid_title.split("_")[-1]
                grid[i].text(0, 1.01, grid_title, ha="left", va="bottom", transform=grid[i].axes.transAxes)

                grid.cbar_axes[i].colorbar(im)
                im.colorbar.formatter = im.colorbar.ax.yaxis.get_major_formatter()
                im.colorbar.ax.get_xaxis().set_ticks([])
                im.colorbar.ax.get_xaxis().set_ticks([], minor=True)
                im.colorbar.ax.yaxis.set_minor_formatter(mticker.LogFormatter())

            grid[i].get_xaxis().set_major_locator(mticker.MaxNLocator(nbins="auto"))
            grid[i].get_yaxis().set_major_locator(mticker.MaxNLocator(nbins="auto"))

            grid[i].get_xaxis().get_major_formatter().set_useOffset(False)
            grid[i].get_yaxis().get_major_formatter().set_useOffset(False)

        self._fig_maps.canvas.draw_idle()

    def _hide_total_count_map_preview(self):
        self._fig_maps.set_visible(False)

    def update_total_count_map_preview(self, *, hide=False, new_plot=False):
        """
        Update total count map preview based on available/selected dataset and `hide` flag.

        Parameters
        ----------
        hide: bool
            `True` - plot data if datasets are available and at least one dataset is selected,
            otherwise hide the plot, `False` - hide the plot in any case
        new_plot: bool
            `True` - plotting new data that was just loaded, reset the plot settings
        """

        if new_plot and not hide:
            # Clear the displayed data range. The range will be computed based on the available data.
            self.clear_map_preview_range()

        # Find out if any data is selected
        show_plot = False
        if self.io_model.data_sets:
            show_plot = any([_.selected_for_preview for _ in self.io_model.data_sets.values()])
        logger.debug(f"LinePlotModel.update_total_count_map_preview(): show_plot={show_plot} hide={hide}")
        if show_plot and not hide:
            logger.debug("LinePlotModel.update_total_count_map_preview(): plotting existing datasets")
            self._show_total_count_map_preview()
        else:
            logger.debug("LinePlotModel.update_total_count_map_preview(): hiding plots")
            self._hide_total_count_map_preview()
        self._fig_maps.canvas.draw()



from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QRadioButton, QButtonGroup, QComboBox
from qtpy.QtCore import Slot

from .useful_widgets import set_tooltip, global_gui_variables
from ..model.lineplot import PlotTypes, EnergyRangePresets

from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)

import logging

logger = logging.getLogger(__name__)


class PreviewPlotSpectrum(QWidget):
    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        self.cb_plot_type = QComboBox()
        self.cb_plot_type.addItems(["LinLog", "Linear"])
        self.cb_plot_type.setCurrentIndex(self.gpc.get_preview_plot_type())
        self.cb_plot_type.currentIndexChanged.connect(self.cb_plot_type_current_index_changed)

        self.rb_selected_region = QRadioButton("Selected region")
        self.rb_selected_region.setChecked(True)
        self.rb_full_spectrum = QRadioButton("Full spectrum")
        if self.gpc.get_preview_energy_range() == EnergyRangePresets.SELECTED_RANGE:
            self.rb_selected_region.setChecked(True)
        elif self.gpc.get_preview_energy_range() == EnergyRangePresets.FULL_SPECTRUM:
            self.rb_full_spectrum.setChecked(True)
        else:
            logger.error(
                "Spectrum preview: incorrect Enum value for energy range was used:\n"
                "    Report the error to the development team."
            )

        self.btn_group_region = QButtonGroup()
        self.btn_group_region.addButton(self.rb_selected_region)
        self.btn_group_region.addButton(self.rb_full_spectrum)
        self.btn_group_region.buttonToggled.connect(self.btn_group_region_button_toggled)

        self.mpl_canvas = FigureCanvas(self.gpc.plot_model._fig_preview)
        self.mpl_toolbar = NavigationToolbar(self.mpl_canvas, self)

        # Keep layout without change when canvas is hidden (invisible)
        sp_retain = self.mpl_canvas.sizePolicy()
        sp_retain.setRetainSizeWhenHidden(True)
        self.mpl_canvas.setSizePolicy(sp_retain)

        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(self.cb_plot_type)
        hbox.addStretch(1)
        hbox.addWidget(self.rb_selected_region)
        hbox.addWidget(self.rb_full_spectrum)
        vbox.addLayout(hbox)

        vbox.addWidget(self.mpl_toolbar)
        vbox.addWidget(self.mpl_canvas)
        self.setLayout(vbox)

        self._set_tooltips()

    def _set_tooltips(self):
        set_tooltip(self.cb_plot_type, "Use <b>Linear</b> or <b>LinLog</b> axes to plot spectra")
        set_tooltip(
            self.rb_selected_region,
            "Plot spectrum in the <b>selected range</b> of energies. The range may be set "
            "in the 'Model' tab. Click the button <b>'Find Automatically ...'</b> "
            "to set the range of energies before finding the emission lines. The range "
            "may be changed in General Settings dialog (button <b>'General ...'</b>) at any time.",
        )
        set_tooltip(self.rb_full_spectrum, "Plot full spectrum over <b>all available eneriges</b>.")

    def update_widget_state(self, condition=None):
        if condition == "tooltips":
            self._set_tooltips()
        self.mpl_toolbar.setVisible(self.gui_vars["show_matplotlib_toolbar"])

        # Hide Matplotlib canvas during computations
        state_compute = global_gui_variables["gui_state"]["running_computations"]
        self.mpl_canvas.setVisible(not state_compute)

    @Slot()
    @Slot(bool)
    def redraw_preview_plot(self):
        # It is assumed that the plot is visible
        self.gpc.update_preview_spectrum_plot()

    def btn_group_region_button_toggled(self, button, checked):
        if checked:
            if button == self.rb_selected_region:
                self.gpc.set_preview_energy_range(EnergyRangePresets.SELECTED_RANGE)
                self.gpc.update_preview_spectrum_plot()
                logger.debug("GUI: Display only selected region")
            elif button == self.rb_full_spectrum:
                self.gpc.set_preview_energy_range(EnergyRangePresets.FULL_SPECTRUM)
                self.gpc.update_preview_spectrum_plot()
                logger.debug("GUI: Display full spectrum")
            else:
                logger.error(
                    "Spectrum preview: unknown button was toggled. "
                    "Please, report the error to the development team."
                )

    def cb_plot_type_current_index_changed(self, index):
        try:
            self.gpc.set_preview_plot_type(PlotTypes(index))
            self.gpc.plot_model.update_preview_spectrum_plot()
        except ValueError:
            logger.error(
                "Spectrum preview: incorrect index for energy range preset was detected.\n"
                "Please report the error to the development team."
            )



import logging
from logging import NullHandler
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

logger = logging.getLogger("pyxrf")
logger.addHandler(NullHandler())



try:
    from databroker.v0 import Broker
except ModuleNotFoundError:
    from databroker import Broker

from hxntools.handlers.xspress3 import Xspress3HDF5Handler
from hxntools.handlers.timepix import TimepixHDF5Handler

db = Broker.named("hxn")
# db_analysis = Broker.named('hxn_analysis')

db.reg.register_handler(Xspress3HDF5Handler.HANDLER_NAME, Xspress3HDF5Handler, overwrite=True)
db.reg.register_handler(TimepixHDF5Handler._handler_name, TimepixHDF5Handler, overwrite=True)



from setuptools import setup, find_packages

import versioneer

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

requirements = ["setuptools"] + requirements

setup(
    name="pyxrf",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Brookhaven National Laboratory",
    url="https://github.com/NSLS-II/PyXRF",
    packages=find_packages(),
    entry_points={"console_scripts": ["pyxrf = pyxrf.pyxrf_run:run"]},
    package_data={"configs": ["*.json"], "pyxrf.core": ["*.yaml"]},
    include_package_data=True,
    install_requires=requirements,
    python_requires=">=3.6",
    license="BSD",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries",
        "Intended Audience :: Science/Research",
    ],
)
